#! /usr/bin/python
"""
Usage:
  test_l1_show.py
  test_l1_show.py [--count NUM] [--golden] [--gss] [--addgolden] [--test LIST] [--debug MODULES]
  test_l1_show.py [--count NUM] [--gss] [--retest DIRNAME] [--addgolden] [--test LIST]  [--debug MODULES]
  test_l1_show.py [--case DIR] [--gss] [--retest DIRNAME] [--addgolden] [--test LIST]  [--debug MODULES]
  test_l1_show.py -h | --help

Options:
  -h, --help                Show this screen
  --count NUM               Number of cl_supports to process before exiting. [default: 200]  Currently the cl_supports are in a random order
  --case DIR                Test just one specified case dir
  -g, --golden              Collect Golden output from baseline l1-show version to compare changes to later
  --gss                     Run on gss stored data
  -r, --retest DIRNAME      Rerun the output json results files back through the comparison. DIRNAME is relative to the results dir
  -t, --test LIST           Run test names that contain string in LIST
  --addgolden               If new data is found or new parsing works better, add the new data to the golden repo
  -v, --debug MODULES       Enable debugging on the list of MODULES. Specify 'main' to debug this module itself.

"""


import os
import sys
import time
import json
import subprocess
import shlex
import logging
from docopt import docopt

from cl_parsers import parse

# This file only runs in cl4, not cl3.x
from l1_show.l1_show_lib import setup_logging, debug_logging, list_logger_levels


def write_results_file(output_text=None, dirname=None, filename=None):
    if dirname is None:
        dirname = './'

    if output_text and filename:
        outname = os.path.join(dirname, filename)
        debug('writing file: {}'.format(outname))
        with open(outname, 'w') as outfile:
            outfile.write(output_text)

    else:
        error('File text or file name was null, no file created')
        return False


def read_local_file(filename, mode='rt'):
    try:
        with open(filename, mode) as f:
            return f.read().strip()
    except IOError:
        info('File: {} not found'.format(filename))
        return None


def read_json_file(json_filename):
    data = read_local_file(json_filename)
    # this pulls out the whole file as one dct, or list.
    # It assumes the calling function parses the struct as needed
    if data:
        return json.loads(data)
    else:
        return None


def shell_out_catch_errors(input_line):
    try:
        process_obj = subprocess.run(shlex.split(input_line), capture_output=True, encoding='UTF-8')
        output = process_obj.stdout
        output_errors = process_obj.stderr
        output_return_code = process_obj.returncode
        if output_errors or output_return_code:
            error('Rtn Code: {} Output errors: {}'.format(output_return_code, output_errors).strip())
        return output, output_errors, output_return_code

        # old: output_json = subprocess.check_output(shlex.split(l1show_command), encoding='UTF-8')
        # or could have appended: .decode('UTF-8') instead of the encoding flag
    except Exception as e:
        error('Command Failed: {}'.format(input_line))
        output_errors ='Unexpected failure: {}'.format(e)
        return None, output_errors, 255


def test_topology_dot(input_dir='./', golden_dirname='./', result_dirname='./', output_file_prefix='',
                      golden=False, addgolden=False, retest_flag=False):

    test_type = 'topology'
    result_dirname = os.path.join(result_dirname, test_type)
    golden_dirname = os.path.join(golden_dirname, test_type)

    if retest_flag:
        retest_file = os.path.join(result_dirname, '{}.json'.format(output_file_prefix))
        info('Retesting json file: {}'.format(retest_file))
        output_dct = read_json_file(retest_file)
    else:
        output_dct = process_topology_dot(input_dir)

    golden_filename = '{}.json'.format(output_file_prefix)
    golden_dct = read_json_file(os.path.join(golden_dirname, golden_filename))

    return compare_dictionaries(input_dir, golden_dirname, result_dirname, test_type, output_file_prefix, output_dct, golden_dct,
                                golden, addgolden, retest_flag)


def test_ptmctl(input_dir='./', golden_dirname='./', result_dirname='./', output_file_prefix='',
                golden=False, addgolden=False, retest_flag=False):

    test_type = 'ptmctl'
    # Compare to golden topo info for now
    result_dirname = os.path.join(result_dirname, test_type)
    golden_dirname = os.path.join(golden_dirname, 'topology')

    if retest_flag:
        retest_file = os.path.join(result_dirname, '{}.json'.format(output_file_prefix))
        info('Retesting json file: {}'.format(retest_file))
        output_dct = read_json_file(retest_file)
    else:
        output_dct = process_ptmctl_detail(input_dir)

    golden_filename = '{}.json'.format(output_file_prefix)
    golden_dct = read_json_file(os.path.join(golden_dirname, golden_filename))
    debug('Golden_dct 1: {}'.format(golden_dct))

    golden_dct = get_relevant_topo_data(input_dir, golden_dct)

    # if golden_dct:
    #     if 'links' in golden_dct:
    #         golden_dct = golden_dct['links']
    #     else:
    #         golden_dct = None

    debug('Golden_dct 3: {}'.format(golden_dct))
    debug('Ptmctl_dct: {}'.format(output_dct))
    return compare_dictionaries(input_dir, golden_dirname, result_dirname, test_type, output_file_prefix, output_dct, golden_dct,
                                golden, addgolden, retest_flag)


def get_relevant_topo_data(dirname, topology_dct):
    new_topo_dct = {}
    filename = os.path.join(dirname, 'etc/hostname')
    if not os.path.isfile(filename):
        info('No hostname file exists in this cl_support')
        return None
    hostname = read_local_file(filename).strip()

    if not topology_dct or 'links' not in topology_dct:
        debug('topology.dot is Null or has no parsed links (edge statements)')
    else:
        # As of 4.3.0/3.7.14, topology_dct inserts a 'links' layer to separate links (edge stmts) from other attributes
        # debug('Golden_dct 2: {}'.format(topology_dct))
        for dutname in topology_dct['links']:
            # Add any lines that are attached to this DUT where this DUT is in the first half of the line
            if dutname.strip("\"") == hostname.split('.')[0]:
                # if dutname not in new_topo_dct:
                #     new_topo_dct[dutname] = {}
                debug('Adding ports for dut: {} in the topology.dot first half data'.format(dutname))
                for port in topology_dct['links'][dutname]:
                    st_port = port.strip('\"')
                    if 'neighbor_port' in topology_dct['links'][dutname][port] and topology_dct['links'][dutname][port]['neighbor_port']:
                        new_topo_dct[st_port] = '{}:{}'.format(
                            topology_dct['links'][dutname][port]['neighbor_dut'].strip('\"'),
                            topology_dct['links'][dutname][port]['neighbor_port'].strip('\"'))
                        # new_topo_dct[st_port]['neighbor_port'] = topology_dct['links'][dutname][port]['neighbor_port'].strip('\"')
                # new_topo_dct.update(topology_dct['links'][dutname])
                continue

            # ensure any local ports on the 2nd half of the link are included in the dictionary keys
            for port in topology_dct['links'][dutname]:
                debug('Searching for {} in the topology 2nd half data'.format(hostname.split('.')[0]))
                neighbor_dut  = topology_dct['links'][dutname][port]['neighbor_dut']
                neighbor_port = topology_dct['links'][dutname][port]['neighbor_port']

                debug('Testing dutname: {}, port: {}. Neighbor DUT: {}, Neighbor port: {}'.format(
                    dutname, port, neighbor_dut, neighbor_port))

                # if (neighbor_dut.strip("\"") == hostname.split('.')[0]) and neighbor_dut not in new_topo_dct:
                #     debug('neighbor_dut matched hostname, adding key: {}'.format(neighbor_dut))
                #     new_topo_dct[neighbor_dut] = {}

                if (neighbor_dut.strip("\"") == hostname.split('.')[0]) and neighbor_port not in new_topo_dct:
                    debug('Adding port {} to local dictionary, with neighbor: {} port: {}'.format(
                        neighbor_port.strip('\"'), dutname.strip('\"'), port.strip('\"')))
                    new_topo_dct[neighbor_port.strip('\"')] = {}
                    new_topo_dct[neighbor_port.strip('\"')] = '{}:{}'.format(dutname.strip('\"'), port.strip('\"'))
                    # new_topo_dct[neighbor_port.strip('\"')]['neighbor_port'] = port.strip('\"')

    return new_topo_dct


def run_l1_show(input_dir='./', golden_dirname='./', result_dirname='./', output_file_prefix='', golden=False,
                addgolden=False, retest_flag=False):

    test_type = 'l1-show'
    golden_exists = ''
    result_dirname = os.path.join(result_dirname, test_type)
    golden_dirname = os.path.join(golden_dirname, test_type)
    golden_filename = '{}.json'.format(output_file_prefix)
    golden_filepath = os.path.join(golden_dirname, golden_filename)

    if retest_flag:
        retest_file = os.path.join(result_dirname, '{}.json'.format(output_file_prefix))
        info('Retesting json file: {}'.format(retest_file))
        output_dct = read_json_file(retest_file)
    else:
        l1show_command = 'l1-show all --cl-support {} --json'.format(input_dir)
        output_json, output_errors, returncode = shell_out_catch_errors(l1show_command)

        if not output_json or int(returncode) != 0:
            if os.path.exists(golden_filepath):
                golden_exists = "FAIL: Golden exists, but "
            error('{}No json output received. Skipping'.format(golden_exists))
            return False
        else:
            debug('mbrown:XXXX output json: {}'.format(output_json))
            output_filename = '{}.json'.format(output_file_prefix)

            write_results_file(output_json, result_dirname, output_filename)
            # Don't need to write file later, just read it back from json here
            output_dct = json.loads(output_json)

    golden_dct = read_json_file(golden_filepath)

    return compare_dictionaries(input_dir, golden_dirname, result_dirname, test_type, output_file_prefix, output_dct, golden_dct,
                                golden, addgolden, retest_flag=True)


def compare_dictionaries(input_dir, golden_dirname, result_dirname, test_type, output_file_prefix, output_dct, golden_dct,
                         golden=False, addgolden=False, retest_flag=False):

    golden_filename = '{}.json'.format(output_file_prefix)
    output_filename = '{}.json'.format(output_file_prefix)

    if golden:
        if output_dct:
            # Store data to file and exit
            info('Parse data was successful, writing Golden: {}'.format(golden_filename))
            golden_output = json.dumps(output_dct, indent=4)
            write_results_file(golden_output, golden_dirname, golden_filename)

    else:
        # This is a test run, not a golden gathering run
        if output_dct:
            # Topo file exists and was properly parsed
            if not retest_flag:
                info('Valid data. Storing new output in {}/{}'.format(result_dirname, output_filename))
                valid_output = json.dumps(output_dct, indent=4)
                write_results_file(valid_output, result_dirname, output_filename)

            if golden_dct:
                # Temp - subtract out 'node_statement' and global_param from golden (old format)
                # golden_dct.pop('node_statement', None)
                # golden_dct.pop('global_param', None)
                # golden_dct.pop('attr_statement', None)
                # # compare the results to the 'links' subtree in output_dct (new format)
                # if 'links' in output_dct:
                #     output_dct = output_dct['links']
                # else:
                #     warn('Topology dot file parsed, but no links were found')

                if output_dct != golden_dct:
                    # Golden and output dct exists but they don't match, this is a full failure
                    # Store differences, or if not possible, entire file
                    error('FAIL: New {} parse does not match Golden: {}'.format(test_type, input_dir))
                    for key in sorted(output_dct.keys()):
                        if output_dct[key] != golden_dct.get(key, {}):
                            warn('First failed key: {}'.format(key))
                            if isinstance(output_dct[key], dict):
                                for key2 in output_dct[key]:
                                    if (output_dct[key][key2] != golden_dct.get(key, {}).get(key2, {})):
                                        warn('First failed key2: {}'.format(key2))
                                        warn('Output: val: {}'.format(output_dct[key][key2]))
                                        warn('Golden: val: {}'.format(golden_dct.get(key, {}).get(key2, {})))
                                        break
                            break

                    # warn('Output keys: {}'.format(sorted(output_dct.keys())))
                    # warn('Golden keys: {}'.format(sorted(golden_dct.keys())))
                    return False
                else:
                    # They both exist and match
                    info('PASS: Golden and parsed {} match'.format(test_type))
                    return True

            else:
                # No golden data exists
                # either something was fixed by this commit, or there was no data before
                # it could be that a new cl_support was added
                warn('Golden did not exist. New {} was parsed: {}'.format(test_type, input_dir))
                if addgolden:
                    info('New Golden output stored in {}'.format(golden_filename))
                    golden_output = json.dumps(output_dct, indent=4)
                    write_results_file(golden_output, golden_dirname, golden_filename)
                return False

        else:
            # No valid output from topo file
            if golden_dct:
                error('FAIL: Golden dictionary exists, but current {} did not parse'.format(test_type))
            else:
                info('No {} parsed nor Golden dictionaries existed to compare'.format(test_type))
            return False


def process_topology_dot(dirname):
    filename = os.path.join(dirname, 'etc/ptm.d/topology.dot')
    if not os.path.isfile(filename):
        info('No topology.dot file exists in this cl_support')
        return None
    output = read_local_file(filename)
    if output:
        topology_dct = parse('cat ' + filename, output)
    else:
        topology_dct = None

    return topology_dct


def process_ptmctl_detail(dirname):
    # TODO:  replace with dut.ptmctl() in future.
    header_strings = ['# Start:', 'Unable to connect to PTMD', 'ERR:', '-------------', 'port   cbl', 'status  nbr', '# Completed']
    ptm_dct = {}
    valid_file = None
    for file_option in ['Support/ptmctl.detail', 'support/ptmctl.detail']:
        filename = os.path.join(dirname, file_option)
        if os.path.isfile(filename):
            valid_file = filename
            break
    if not valid_file:
        info('No ptmctl.detail file exists in this cl_support')
        return None
    else:
        debug('Valid ptmctl.detail file found in dir: {}'.format(dirname))

    output = read_local_file(filename)
    if output:
        for line in output.strip().split('\n'):
            # debug('XXX Line: {}'.format(line.strip()))
            line = line.strip()
            if line:
                hdr = False
                for header_line in header_strings:
                    # debug('XXXX Testing {} against {}'.format(line, header_line))
                    if line.startswith(header_line):
                        # debug('XXXXX Matches')
                        hdr = True
                        break
                if not hdr and (line.startswith('sw') or line.startswith('eth')):
                    # debug('XXXX Line: {}'.format(line))
                    ptm_line = line.split()
                    if ptm_line[2]:
                        ptm_dct[ptm_line[0]] = ptm_line[2]
                        debug('Added ptm entry: {}: {}'.format(ptm_line[0], ptm_line[2]))

    return ptm_dct


def relink_file_if_new_simple(filename, filelink):
    #  this only works if this script is executing on a host that IS the www server or that is NFS mounted to the
    #  same /home tree...
    # This was originally done so that I didn't have to deal with permissions on monster-16:/var/www/image
    target_file = os.path.abspath(filename)
    debug('abspath to target file resolves to {}'.format(target_file))
    if os.path.lexists(filelink):
        # testbed_db_file_path=testbed_data_file_path(results)
        linked_file = os.path.abspath(os.path.join(os.path.dirname(filelink), os.readlink(filelink)))
        debug('abspath link to previous linked file resolves to {}'.format(linked_file))
        if linked_file == target_file:
            info('File matches previous file - no change.')
            return False
        else:
            info('File doesnt match previous file, remove old link and relink to new file')
            os.remove(filelink)
            debug('Linking {0} -> {1}'.format(filelink, target_file))
            os.symlink(target_file, filelink)
            return True
    else:
        info('No previous link existed, Creating link to new file')
        os.symlink(target_file, filelink)
        return True


def generate_clsupport_list(rootdir, maxcount=100000):
    dirlist = []
    count = 0
    count2 = 0
    count3 = 0
    with os.scandir(rootdir) as it:
        for entry in it:
            if entry.is_dir() and count3 < maxcount:
                count += 1
                with os.scandir(os.path.join(rootdir, entry.name)) as it2:
                    for entry1 in it2:
                        if entry1.is_dir() and count3 < maxcount:
                            count2 += 1
                            with os.scandir(os.path.join(rootdir, entry.name, entry1.name)) as it3:
                                for entry2 in it3:
                                    if entry2.is_dir() and entry2.name.startswith('cl_support') and count3 < maxcount:
                                        count3 += 1
                                        dirlist.append(os.path.join(rootdir, entry.name, entry1.name, entry2.name))

    info('cl_support counts: Customer: {}, Cases: {}, cl_supports: {}'.format(count, count2, count3))
    return dirlist


def split_dirname(dirName):
    this_support = dirName.split('/')[-1]
    case = dirName.split('/')[-2].split('_')[0]
    customer = dirName.split('/')[-3].split('_')[0]

    return this_support, case, customer


def run_tests(dirname, output_file_prefix):
    if not arguments['--test']:
        run_all = True
    else:
        run_all = False
    for testid in arguments['--test'].split(','):
        if testid in 'test_topology_dot' or run_all:
            test_topology_dot(dirname, golden_dir, result_dir, output_file_prefix, arguments['--golden'],
                              arguments['--addgolden'], retest)

        if testid in 'run_l1_show' or run_all:
            run_l1_show(dirname, golden_dir, result_dir, output_file_prefix, arguments['--golden'],
                        arguments['--addgolden'], retest)

        if testid in 'test_ptmctl' or run_all:
            test_ptmctl(dirname, golden_dir, result_dir, output_file_prefix, arguments['--golden'],
                        arguments['--addgolden'], retest)


def main():

    # Number of cl-supports to process
    if arguments['--count'] == 'all':
        maxcount = 100000
    else:
        maxcount = int(arguments['--count'])
    info('Processing max count of cl_supports: {}'.format(maxcount))

    if arguments['--debug']:
        debug_logging(arguments['--debug'].split(','), this_log)
        list_logger_levels()

    # l1show_command = 'l1-show all --cl-support'
    # cmd_list = [l1show_command]

    # Name of a cl-support file so we know when we've hit a new support directory
    this_support = 'abcdefg'

    if arguments['--case']:
        dirname = arguments['--case']
        this_support, case, customer = split_dirname(dirname)
        info_line = 'Analyzing: {}/{}/{}'.format(customer, case, this_support)
        info(info_line)
        # Topology.dot processing
        output_file_prefix = '{}_{}_{}'.format(customer, case, this_support)

        run_tests(dirname, output_file_prefix)

    else:
        # Loops through the root directory of all the cl-supports looking for extracted support files
        # for dirName, subDirList, fileList in os.walk(rootdir):
        for dirname in generate_clsupport_list(rootdir, maxcount):
            this_support, case, customer = split_dirname(dirname)
            # Future:  Use the l1_show Dut for processing ptmctl and other items
            # dut = Dut(dirName)
            info_line = 'Analyzing: {}/{}/{}'.format(customer, case, this_support)
            info(info_line)
            # Topology.dot processing
            output_file_prefix = '{}_{}_{}'.format(customer, case, this_support)

            run_tests(dirname, output_file_prefix)


if __name__ == '__main__':

    this_log = logging.getLogger(__name__)
    debug = this_log.debug
    info = this_log.info
    warn = this_log.warning
    error = this_log.error

    arguments = docopt(__doc__, version='test_l1_show 0.9')
    print(arguments)

    if arguments['--gss']:
        # The root directory for customer cl-supports on gssvm2
        rootdir = '/gss/case_files/'
        log_dir = '/home/mbrown/'
        golden_dir = "/home/mbrown/l1_show_golden"
        results_dir = "/home/mbrown/l1_show_results"
    else:
        # The root directory for customer cl-supports on laptop
        rootdir = "/Users/mbrown/Customer/"
        log_dir = "/Users/mbrown/Customer/"
        golden_dir = "/Users/mbrown/Customer/l1_show_golden"
        results_dir = "/Users/mbrown/Customer/l1_show_results"

    test_types = ['topology', 'l1-show', 'ptmctl']

    datestamp = time.strftime("%Y-%m-%d_%H-%M-%S")
    if arguments['--retest']:
        result_dir = os.path.join(results_dir, arguments['--retest'])
        log_dir = os.path.join(result_dir, 'logs-retest-{}'.format(datestamp))
        retest = True
    else:
        retest = False
        result_dir = os.path.join(results_dir, 'results-' + datestamp)
        os.makedirs(result_dir)
        os.makedirs(os.path.join(result_dir, 'logs'))
        for test_type_dir in test_types:
            os.makedirs(os.path.join(result_dir, test_type_dir), exist_ok=True)
            os.makedirs(os.path.join(golden_dir, test_type_dir), exist_ok=True)
        relink_file_if_new_simple(result_dir, os.path.join(results_dir, 'latest_results'))

    # default_level = logging.INFO
    # logformat = '%(asctime)s: %(name)20s: %(levelname)s: %(message)s'
    print('config_file_path={}'.format(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'default_log_config.json')))
    setup_logging(config_file_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'default_log_config.json'),
                  log_file_dir=os.path.join(result_dir, 'logs'))
    # setup_logging(default_level=default_level, quiet=True, logformat=logformat)
    # logging.basicConfig(level=default_level, format=logformat)

    sys.exit(main())


