#!/usr/bin/env -S python -i

# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import re

from l1_show.l1_show_lib import *

this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error


class Base(object):

    def __init__(self, path_prefix='/'):
        self.path_prefix = path_prefix
        self.support_dir = self.get_support_dir()

        self.fqdn = self.get_hostname()
        self.hostname = self.fqdn.split('.')[0]
        self.platform_object = self.get_platform_object()
        self.asic_vendor = None
        self._get_asic_vendor()
        self.asic = None
        self._get_asic()
        self.pcs_counters_dct = {}
        self.linux_to_sdk = {}

        self.ethmode_map = {
            '40g-cr4':    24,
            '40g-sr4':    25,
            '40g-lr4':    26,
            '10g-cr':     42,
            '10g-sr':     43,
            '10g-lr':     44,
            '10g-lrm':    45,
            '100g-sr4':   37,
            '100g-cr4':   38,
            '100g-lr4':   39,
            '25g-cr':     31,
            '25g-sr':     33,
            '10g-t':      12,
            '10g-tsr':    12,
            '50g-cr2':    34,
            '50g-sr2':    40,
            '2.5g-t':     47,
            '5g-t':       48,
            '1000base-t':  5,
            '1000base-x': 41,
            'Zero':        0,
        }
        self.reverse_ethmode_map = {}
        for ethmode_text in sorted(self.ethmode_map):
            if self.ethmode_map[ethmode_text] not in self.reverse_ethmode_map:
                self.reverse_ethmode_map[str(self.ethmode_map[ethmode_text])] = ethmode_text

    def mlxlink(self, port_list):
        if not isinstance(port_list, list):
            port_list = [port_list]

        mlxlink_base = (
            # Space intentionally missing after extra_flag for cl_support parsing
            "{mlxlink_cmd} -d {pci_name} -c -m -e {extra_flag}--show_fec " +
            "-p {mlx_port} {post_cmd}")

        if self.asic_vendor != 'mlx':
            return None

        mlxlink_dct = {}
        for port in port_list:
            mlx_port = mlxlink_port(port)
            mlxlink_cmd = self.format_mlxlink_cmd(mlxlink_base, mlx_port)
            mlxlink_dct[port] = self.get_mlxlink_data(mlxlink_cmd)
            if not mlxlink_dct[port]:
                debug('mlxlink parser returned None for port {}'.format(port))

        dut_obj_install_dict(self, 'mlxlink_dct', mlxlink_dct)

    def _get_asic_vendor(self):
        if self.platform_object and self.platform_object.switch and self.platform_object.switch.chip:
            self.asic_vendor = self.platform_object.switch.chip.sw_base
        else:
            warn('ASIC vendor not found on this platform')
            self.asic_vendor = None

    def _get_asic(self):
        if self.platform_object and self.platform_object.chip and self.platform_object.chip.chip_dir_list:
            for platform_dir in self.platform_object.chip.chip_dir_list:
                dir_list = platform_dir.strip().split('/')
                for dir_item in dir_list:
                    if dir_item != '' and self.asic_vendor not in dir_item:
                        self.asic = dir_item
                        return
        warn('ASIC not found on this platform')
