# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import os
import sys
import re
import glob
import logging
import xml.etree.ElementTree as ET

try:
    from collections import namedtuple as NamedTuple
except ImportError:
    from typing import NamedTuple

from cl_parsers.parse import parse

from l1_show.base import Base
from l1_show.l1_show_lib import *

this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error

# See cl-parsers/.../bcm_phy_diag_dsc.phy
# Capture groups are used in order to include this text in the parsing as well
phy_patterns = [
    re.compile(r'^\s*(TSC\-A.*)$', re.MULTILINE),
    re.compile(r'^\s*(tsc[ef16]+_phy_pmd_info_dump.*)$', re.MULTILINE),
    re.compile(r'^\s*(blackhawk_phy_pmd_info_dump:[0-9]+.*)$', re.MULTILINE),
    re.compile(r'^\s*(DSC parameters for port [0-9]+.*?)$', re.MULTILINE)
]


def critical(mesg):
    this_log.critical(mesg)
    this_log.critical('Exiting abnormally')
    sys.exit(1)


def phy_parse_port(port):
    pre, num = re.match(r'^(\D+)(\d+)$', port).groups()
    return pre, int(num)


def parse_command_output_file(filename, command=None, match=1):
    output = read_local_file(filename)
    if output is None:
        return None

    if command is None:
        command = r'.*?'

    matches = re.findall(
        r'^# Start: ({})\s*$((?:.|[\r\n])+?)^# Completed in [\d.]+ seconds: \1'.format(command),
        output,
        re.MULTILINE
    )
    return matches[match - 1][1].strip() if len(matches) >= match else None


def get_bcm_chip_dir_list(dut):
    fn = os.path.join(dut.path_prefix, dut.support_dir, 'show.params')
    output = parse_command_output_file(fn)
    if output is None:
        chip = 'N/A'
    else:
        # TODO - if a bash error occurred, this will not match 1st line, use re.search and test
        chip = re.match(r'^driver .* \((.*)\)$', output, re.MULTILINE).group(1)

    # Performs some common transformations
    chip = chip.capitalize()
    if chip == 'Apache':
        chip = 'Maverick'
    elif chip == 'Trident2':
        chip = 'TridentTwo'
    elif chip == "Tomahawk_plus":
        chip = "TomahawkPlus"

    return ['/bcm/' + chip, '/bcm']


class Cl_Support(Base):

    def __init__(self, cl_support):
        Base.__init__(self, cl_support)
        self.log_throttle = []

    def get_hostname(self):
        fn = os.path.join(self.path_prefix, self.support_dir, 'systemd.status')
        if not os.path.isfile(fn):
            return 'N/A'
        with open(fn, 'r') as f:
            # Skips first line, then gets second to get hostname
            next(f)
            line = next(f)

        # Parses line to get hostname
        return re.search('\s+(\S+)', line).group(1)

    def get_support_dir(self):
        for dir_option in ['Support', 'support']:
            dirname = os.path.join(self.path_prefix, dir_option)
            if os.path.isdir(dirname):
                if dir_option == 'support':
                    critical('CL 2.x not supported in this tool')
                else:
                    debug('Valid Support dir found in dir: {}'.format(dirname))
                    return dir_option

        critical('No Support/support dir exists in this cl_support')
        return None


    def get_default_fec_value(self):
        return 'N/A'

    def get_local_portwd_conf(self):
        fn = os.path.join(self.path_prefix, 'etc/cumulus/portwd.conf')
        if not os.path.isfile(fn):
            return None
        else:
            return fn

    def get_platform_object(self):
        # NamedTuple is used to mimic class-structure that probe() creates
        Switch_Chip = NamedTuple('Switch_Chip', 'sw_base')
        Switch = NamedTuple('Switch', 'chip')
        Chip = NamedTuple('Chip', 'chip_dir_list')
        Object = NamedTuple('Object', 'name switch chip')

        filename = os.path.join(self.path_prefix, self.support_dir, 'platform.detect')
        if not os.path.isfile(filename):
            name = 'N/A'
        else:
            with open(filename, 'r') as f:
                # Skips first line, then gets name
                next(f)
                name = next(f).strip()

        # Currently, we assume other info based on presence of specific files
        if 'mlnx' in name:
            switch_chip = 'mlx'
            chip_dir_list = ['/mlx/Spectrum', '/mlx']
        else:
            switch_chip = 'bcm'
            chip_dir_list = get_bcm_chip_dir_list(self)

        return Object(
            name,
            Switch(
                Switch_Chip(switch_chip)
            ),
            Chip(chip_dir_list)
        )

    def bcmcmd(self, num, cmd):
        if cmd == 'portstat':
            filename = os.path.join(self.path_prefix, self.support_dir, 'portstat')
        elif cmd == 'port all':
            filename = os.path.join(self.path_prefix, self.support_dir, 'port.all')
        else:
            print('Error: invalid bcmcmd "' + str(cmd) + '"')
            sys.exit(1)

        # Cuts off first and last lines since they are irrelevant
        output = parse_command_output_file(filename)
        if output is None:
            return {}
        cmd = format_cmd('/usr/lib/cumulus/bcmcmd {}: {}'.format(num, cmd))

        parsed_bcm = parse(cmd, output)
        if parsed_bcm is None:
            return {}
        else:
            return parsed_bcm

    # Parses dsc file - since it contains all outputs for a particular sdk,
    # it needs to be filtered down to just the port we want
    # Assumes that the dsc file output is ordered numerically by port
    def bcmcmd_phy(self, port, unit, cmd):
        sdk = phy_parse_port(port)[0]

        fn = os.path.join(self.path_prefix, self.support_dir, 'phy.diag.{}.dsc'.format(sdk))
        output = parse_command_output_file(fn)
        if output is None:
            debug('No file \'{}\' - is CL version > 3.7.10?'.format(fn))
            return {}

        if output == '':
            # Sometimes, data is not recorded by this command. In this case, all
            # we can do is report that and move on
            debug('No matching phy pattern for {} {}'.format(port, unit))
            return parse(cmd, '')

        if not hasattr(self, 'porttab_dct'):
            dut_install_porttab(self)

        # First, filter down to matching sdks
        porttab_dct = {k: v for k, v in self.porttab_dct.items() if re.match(sdk, v[0]) is not None}
        # Next, find relative position
        number = 0
        for k in sorted(porttab_dct.values(), key=lambda x: (
            int(re.match(r'unit(\d+)', x[1]).group(1)),
            int(re.match(sdk + r'(\d+)', x[0]).group(1))
        )):
            if k[0] == port and k[1] == unit:
                break
            number += 1

        match = None
        for p in phy_patterns:
            if p.search(output) is not None:
                match = p
        if match is None:
            critical('No matching phy pattern for {} {}'.format(port, unit))

        # Splits outputs into groups
        # Iterate in pairs - https://stackoverflow.com/a/48347320
        it = iter(match.split(output))
        # Skips first match, which will always be empty - https://docs.python.org/3/library/re.html#re.split
        next(it)
        texts = []
        for m in list(zip(it, it)):
            texts.append(m[0] + m[1])

        debug('dsc output: {}'.format(output))
        return parse(cmd, format_output(texts[number]))

    def ethtool(self, port, flag=None, parse_output=True):
        # cl-support only stores output for first lane, since it's the same for all of them
        groups = re.match(r'(swp[\d]+s)[\d]+', port)
        if groups is not None:
            port = groups.group(1) + '0'

        if flag is None:
            fn = 'ethtool.link'
            cmd_pattern = 'ethtool {}'.format(port)
            cmd = cmd_pattern
        elif flag == '-m':
            fn = 'ethtool.module'
            cmd_pattern = 'ethtool -m {}'.format(port)
            cmd = '{} 2> /dev/null'.format(cmd_pattern)
        elif flag == '-m hex on':
            fn = 'ethtool.module-hex'
            cmd_pattern = 'ethtool -m {} hex on'.format(port)
            cmd = '{} 2> /dev/null'.format(cmd_pattern)
        elif flag == '--show-fec':
            fn = 'ethtool.fec'
            cmd_pattern = 'ethtool --show-fec {}'.format(port)
            cmd = cmd_pattern
        else:
            critical('Unrecognized ethtool flag: {}'.format(flag))

        filename = os.path.join(self.path_prefix, self.support_dir, fn)
        output = parse_command_output_file(filename, cmd_pattern)
        if not output:
            msg = '{} not available in cl_support'.format(fn)
            if msg not in self.log_throttle:
                error('{} not available in cl_support'.format(fn))
                self.log_throttle.append(msg)
            return {}
        if not parse_output:
            return output
        return parse(format_cmd(cmd), format_output(output))

    def lldpctl(self):
        filename = os.path.join(self.path_prefix, self.support_dir, 'lldpctl')
        data = parse_command_output_file(filename)
        if data is None:
            return {}
        dct = parse('lldpctl', data)

        return {
            v['Interface']: {
                'neighbor_port': v.get('PortDescr', 'N/A'),
                'neighbor_switch': v.get('SysName', 'N/A'),
            } for v in dct.values()
        }

    def ptmctl(self):
        filename = os.path.join(self.path_prefix, self.support_dir,  'ptmctl.detail')
        if not os.path.isfile(filename):
            debug('No {} file exists in this cl_support'.format(()))
            return None
        else:
            debug('Valid ptmctl.detail file found in dir: {}/{}'.format(self.path_prefix, self.support_dir))
        data = parse_command_output_file(filename)
        if data is None:
            return {}
        return self.process_ptmctl_detail(data)

    def process_ptmctl_detail(self, data):
        header_strings = ['Unable to connect to PTMD', 'ERR:', '-------------', 'port   cbl', 'status  nbr']
        ptmctl_dct = {}

        if data:
            for line in data.strip().split('\n'):
                # debug('XXX Line: {}'.format(line.strip()))
                line = line.strip()
                if line:
                    hdr = False
                    for header_line in header_strings:
                        # debug('XXXX Testing {} against {}'.format(line, header_line))
                        if line.startswith(header_line):
                            # debug('XXXXX Matches')
                            hdr = True
                            break
                    if not hdr and (line.startswith('sw') or line.startswith('eth')):
                        # debug('XXXX Line: {}'.format(line))
                        ptm_line = line.split()
                        if ptm_line[2]:
                            ptmctl_dct[ptm_line[0]] = ptm_line[2]
                            debug('Added ptm entry: {}: {}'.format(ptm_line[0], ptm_line[2]))

        return ptmctl_dct

    def ip_link_show(self):
        filename = os.path.join(self.path_prefix, self.support_dir, 'ip.addr')
        out = parse_command_output_file(filename)
        if out is None:
            return {}
        return parse('ip link show', format_output(out))

    def get_eeprom_list(self):
        eeprom_list = []
        pattern = re.compile(r'^swp(\d+).*$')
        filename = os.path.join(self.path_prefix, 'var/lib/cumulus/sfptab')
        if not os.path.isfile(filename):
            # This file is not available on shorter/broken cl_supports
            # Use another method to get this if this one fails - ip link keys.
            # This will also give Fixed ports, but better than nothing
            error('var/lib/cumulus/sfptab not found.  cl_support is incomplete. Fixed copper ports will not be marked')
            matches = self.ip_link_show()
            for swp in matches:
                mo = pattern.search(swp)
                if mo:
                    eeprom_list.append(mo.group(1))

        else:
            f = open(filename, 'r')
            for swp in f.readlines():
                mo = pattern.search(swp)
                if mo:
                    eeprom_list.append(mo.group(1))

        debug('eeprom_list: {}'.format(eeprom_list))
        return eeprom_list

    def get_porttab(self):
        filename = os.path.join(self.path_prefix, self.support_dir, 'portmap')
        if not os.path.isfile(filename):
            critical('This cl_support file is a partial capture from a crash and does not contain required l1-show files. Please retry on a full cl_support')
            return {}
        with open(filename, 'r') as f:
            lines = f.readlines()

        pattern = re.compile(r'^(\w+)\s*?(\w+)\s*?(\d+)')

        return {
            match[0]: (match[1], 'unit' + match[2])
            for l in lines for match in pattern.findall(l)
        }

    def get_pci_name(self):
        """
        get_pci_name used to get the output from mlxfwmanager, but matching on mlxlink
         was difficult because of multiple cmd changes.
         Since is the cl-support, we can just grab the pci info
         from the executed cmd in mlxlink file
        """
        fn = os.path.join(self.path_prefix, self.support_dir, 'mlxlink')
        pattern = '# Start: .* -d\s(.*) -p.*'
        output = read_local_file(fn)
        if output is None:
            return None
        output_lines = output.split('\n')
        for line in output_lines:
            if '# Start:' in line:
                mlxlink_match = re.match(pattern, line)
                if mlxlink_match:
                    return mlxlink_match.group(1)

        return 'N/A'

    def format_mlxlink_cmd(self, base, port):
        #  get_pci_name now returns a specific substring from mlxlink output,
        #  rather than from mlxfwmanager output
        return base.format(
            mlxlink_cmd='mlxlink',
            pci_name=self.get_pci_name(),
            extra_flag='',
            mlx_port=port,
            post_cmd='',
        )

    def get_mlxlink_data(self, cmd):
        prefix = 'sh -c '

        fn = os.path.join(self.path_prefix, self.support_dir, 'mlxlink')
        output = read_local_file(fn)
        if output is None:
            return {}
        # TODO: need to not escape any .* or [a-z] brackets
        start = r'# Start: ' + re.escape(prefix + cmd) + r'.*'
        end = r'[\d.]+ seconds: ' + re.escape(prefix + cmd) + r'.*'

        pattern = '(' + end + ')?\n' + start

        mlxlink_out_list = output.split('# Completed in ')
        for link in mlxlink_out_list:
            if re.search(start, link):
                mlxlink_out = re.sub(pattern, '', link)
                return parse(cmd, format_output(mlxlink_out))
                break

    def install_linux_to_mlx(self):
        error('PCS error collection not supported on cl_support files')
        return

    def install_pcs_data(self, port_list, clear=False):
        error('PCS error collection not supported on cl_support files')
        return
