# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import struct
import io
import errno
import os
import logging
import time

logger = None
logs_throttled = {}

logger = logging.getLogger("eeprom-utils")

def encode(string):
    # Python2 strings are implicitly bytes, while in python3 they are
    # implicitly unicode
    try:
        # Why latin-1? https://github.com/SecureAuthCorp/impacket/issues/535
        return string.encode('latin-1')
    except UnicodeDecodeError:
        return str(string)

def set_logger(handle):
    global logger
    logger = logging.getLogger(handle)

def log_throttle(msg, once=False, level='warning'):
    send_log = True
    if msg in logs_throttled.keys():
        throttled = logs_throttled[msg]
        if once:
            send_log = False
        elif throttled.expired:
            throttled = LogMessage(msg)
        else:
            send_log = throttled.repeat()
    else:
        throttled = LogMessage(msg)

    logs_throttled[msg] = throttled

    message = throttled.message

    if send_log:
        if level == 'warning':
            logger.warning(message)
        elif level == 'error':
            logger.error(message)
        elif level == 'info':
            logger.info(message)
        elif level == 'debug':
            logger.debug(message)

    return message

class LogMessage(object):
    '''
    A throttled log message.  Messages are logged with exponential backoff.
    '''
    def __init__(self, msg, expire=60*60, ts=None):
        self.msg = msg
        if ts:
            self.ts = ts
        else:
            self.ts = time.time()

        self.expire = self.ts + expire

        self._seq = 1
        self._repeat = 0

    def repeat(self):
        now = time.time()

        self._repeat += 1

        if now > self.ts + (2 ** self._seq):
            self._seq += 1
            return self.message

        return None

    @property
    def message(self):
        if self._repeat == 0:
            return self.msg
        else:
            return self.msg + ' (repeated %u times)' % self._repeat

    @property
    def expired(self):
        return time.time() > self.expire

class InvalidTransceiverCode(Exception):
    pass

class EEPROMTimeout(Exception):
    pass

class SFFDecoder(object):
    ETH_EXTENDED_CODE_MAP = {  1  : '100g-active',     # 100G AOC or 25G AUI C2M AOC (active optical cable,BER = 5x10^-5)
                               2  : '100g-sr4',        # 100G SR4 or 25G SR
                               3  : '100g-lr4',        # 100G LR4 or 25G LR
                               4  : '100g-er4',        # 100G ER4
                               5  : '100g-sr10',       # 100G SR10
                               6  : '100g-cwdm4fec',   # 100G CWDM4 MSA with FEC
                               7  : '100g-psm4',       # 100G PSM4 Parallel SMF
                               8  : '100g-active-cu',  # 100G ACC or 25G AUI C2M ACC (active copper cable, BER = 5x10^-5)
                               9  : '100g-cwdm4',      # 100G CWDM4 MSA without FEC
                               11 : '100g-cr4',        # 100G CR4 or 25G CR CA-L
                               12 : '25g-cr-ca-s',     # 25G CR CA-S
                               13 : '25g-cr-ca-n',     # 25G CR CA-N
                               16 : '40g-er4',         # 40G ER4
                               17 : '4x10g-sr',        # 4 x 10G SR
                               18 : '40g-psm4',        # 40G PSM4 Parallel SMF
                               19 : 'p1i1-2d1',        # G959.1 profile P1I1-2D1 (10709 MBd, 2km, 1310nm SM)
                               20 : 'p1s1-2d2',        # G959.1 profile P1S1-2D2 (10709 MBd, 40km, 1550nm SM)
                               21 : 'p1l1-2d2',        # G959.1 profile P1L1-2D2 (10709 MBd, 80km, 1550nm SM)
                               22 : '10g-t',           # 10GBASE-T with SFI electrical interface
                               23 : '100g-clr4',       # 100G CLR4
                               24 : '100g-active',     # 100G AOC or 25G AUI C2M AOC (active optical cable,BER < 10^-12)
                               25 : '100g-active-cu',  # 100G ACC or 25G AUI C2M ACC (active copper cable, BER < 10^-12)
                               26 : '100g-dwdm2',      # 100GE-DWDM2 (2 wavelengths on 1550nm DWDM grid, Reach <80km)
                               27 : '100g-wdm',        # 100G 1550nm WDM (4 wavelengths)
                               28 : '10g-tsr',         # 10GBASE-T Short Reach (30 meters)
                               29 : '5g-t',            # 5GBASE-T
                               30 : '2.5g-t',          # 2.5GBASE-T
                               31 : '40g-swdm4',       # 40G SWDM4
                               32 : '100g-swdm4',      # 100G SWDM4
                               33 : '100g-pam4-bidi',  # 100G PAM4 BiDi
                               34 : '4wdm-10-msa',     # 4WDM-10 MSA (10km version of 100G CWDM4 with same RS(528,514) FEC in host system)
                               35 : '4wdm-20-msa',     # 4WDM-20 MSA (20km version of 100GBASE-LR4 with RS(528,514) FEC in host system)
                               36 : '4wdm-40-msa',     # 4WDM-40 MSA (40km reach with APD receiver and RS(528,514) FEC in host system)
                               37 : '100g-dr',         # 100GBASE-DR (clause 140), CAUI-4 (no FEC)
                               38 : '100g-fr-nofec',   # 100G-FR or 100GBASE-FR1 (clause 140), CAUI-4 (no FEC)
                               39 : '100g-lr-nofec',   # 100G-LR or 100GBASE-LR1 (clause 140), CAUI-4 (no FEC)
                               40 : '200g-acc1',       # Active Copper Cable with 50GAUI, 100GAUI-2 or 200GAUI-4 C2M. Providing a worst BER of 10-6 or below
                               49 : '200g-aoc1',       # Active Optical Cable with 50GAUI, 100GAUI-2 or 200GAUI-4 C2M. Providing a worst BER of 10-6 or below
                               50 : '200g-acc2',       # Active Copper Cable with 50GAUI, 100GAUI-2 or 200GAUI-4 C2M. Providing a worst BER of 2.6x10-4 for ACC, 10-5 for AUI, or below
                               51 : '200g-aoc2',       # Active Optical Cable with 50GAUI, 100GAUI-2 or 200GAUI-4 C2M. Providing a worst BER of 2.6x10-4 for ACC, 10-5 for AUI, or below
                               64 : '200g-cr4',        # 50GBASE-CR, 100GBASE-CR2, or 200GBASE-CR4
                               65 : '200g-sr4',        # 50GBASE-SR, 100GBASE-SR2, or 200GBASE-SR4
                               66 : '200g-dr4',        # 50GBASE-FR or 200GBASE-DR4
                               67 : '200g-fr4',        # 200GBASE-FR4
                               68 : '200g-psm4',       # 200G 1550 nm PSM4
                               69 : '50g-lr',          # 50GBASE-LR
                               70 : '200g-lr4',        # 200GBASE-LR4
                               71 : '64g-fc-ea',       # 64GFC EA
                               81 : '64g-fc-sw',       # 64GFC SW
                               82 : '64g-fc-lw',       # 64GFC LW
                               83 : '128g-fc-ea',      # 128GFC EA
                               84 : '128g-fc-sw',      # 128GFC SW
                               85 : '128g-fc-lw',      # 128GFC LW
    # 10 : reserved
    # 14-15 : reserved
    # 40-47 : reserved
    # 52-63 : reserved
    # 71-79 : reserved
    # 86-255 : reserved
    }

    CONNECTOR_CODE_MAP = { 0x00: 'unknown',
                           0x01: 'sc',
                           0x02: 'fibre-channel-style1-copper',
                           0x03: 'fibre-channel-style2-copper',
                           0x04: 'bnc-or-tnc',
                           0x05: 'fibre-channel-coax-hdr',
                           0x06: 'fiber-jack',
                           0x07: 'lc',
                           0x08: 'mt-rj',
                           0x09: 'multiple-optical',
                           0x0a: 'sg',
                           0x0b: 'optical-pigtail',
                           0x0c: 'mpo-1x12',
                           0x0d: 'mpo-2x16',
                           0x20: 'hssdc2',
                           0x21: 'copper-pigtail',
                           0x22: 'rj45',
                           0x23: 'no-separable-connector',
                           0x24: 'mxc-2x16',
                           0x25: 'cs',
                           0x26: 'mini-cs',
                           0x27: 'mpo-2x12',
                           0x28: 'mpo-1x16'
                           }

    @classmethod
    def decode_connector(self, eeprom):
        data = eeprom[self.ADDR_CONNECTOR_CODE]
        if len(data) != 1:
            return None

        concode = struct.unpack('B', data)[0]
        return self.CONNECTOR_CODE_MAP.get(concode, 'unknown')

    @classmethod
    def decode_lpoverride(self, eeprom):
        if not self.ADDR_POWER_CTL:
            return None

        data = eeprom[self.ADDR_POWER_CTL]
        lpov = struct.unpack('B', data)[0] & self.POWER_CTL_LPOVERRIDE_MASK

        return self.POWER_CTL_LPOVERRIDE_MAP.get(lpov, 'unknown')

    @classmethod
    def encode_lpoverride(self, eeprom, value):
        if not self.ADDR_POWER_CTL:
            return

        if value not in self.POWER_CTL_LPOVERRIDE_RMAP:
            raise RuntimeError('invalid lpoverride value: %s' % value)
            return

        data = eeprom[self.ADDR_POWER_CTL]
        lpov = struct.unpack('B', data)[0] & self.POWER_CTL_LPOVERRIDE_MASK
        lpov &= ~self.POWER_CTL_LPOVERRIDE_MASK
        lpov |= self.POWER_CTL_LPOVERRIDE_RMAP[value]

        eeprom[self.ADDR_POWER_CTL] = struct.pack('B', lpov)

    @classmethod
    def enable_high_power(self):
        pass

    @classmethod
    def decode_cdr_present(self, eeprom):
        if not self.ADDR_EXT_IDENT:
            return False

        data = (struct.unpack('B', eeprom[self.ADDR_EXT_IDENT]))[0]
        if (data & self.CDR_EXT_PRESENT_MASK) == self.CDR_EXT_PRESENT_MASK:
            return True
        else:
            return False

    @classmethod
    def decode_cdr_controllable(self, eeprom):
        if not self.ADDR_CDR_CTRL_IMPLMTD:
            return False

        data = (struct.unpack('B', eeprom[self.ADDR_CDR_CTRL_IMPLMTD]))[0]
        if (data & self.CDR_CTRL_IMPLMTD_MASK) == self.CDR_CTRL_IMPLMTD_MASK:
            return True
        else:
            return False

    @classmethod
    def decode_cdr_setting(self, eeprom):
        if not self.ADDR_CDR_CTRL:
            return None

        curr_cdr = (struct.unpack('B', eeprom[self.ADDR_CDR_CTRL]))[0]
        return self.CDR_SEL_MAP.get(curr_cdr)


    @classmethod
    def encode_cdr_setting(self, eeprom, value):
        if not self.ADDR_CDR_CTRL:
            raise NotImplementedError

        if value not in self.CDR_SEL_RMAP:
            return

        desired_cdr = self.CDR_SEL_RMAP.get(value)
        eeprom[self.ADDR_CDR_CTRL] = struct.pack('B', desired_cdr)


    @classmethod
    def decode_tx_disable_implmtd(self, eeprom):
        if not self.ADDR_TX_DISABLE_IMPLMTD:
            return False

        data = (struct.unpack('B', eeprom[self.ADDR_TX_DISABLE_IMPLMTD]))[0]
        if (data & self.TX_DISABLE_IMPLMTD_MASK) == self.TX_DISABLE_IMPLMTD_MASK:
            return True
        else:
            return False

    @classmethod
    def encode_tx_disable(self, eeprom, value):
        if not self.ADDR_TX_DISABLE:
            raise NotImplementedError

        if value not in self.TX_DISABLE_MAP:
            return

        eeprom[self.ADDR_TX_DISABLE] = struct.pack('B', self.TX_DISABLE_MAP.get(value))

    @classmethod
    def decode_bit_rate(self, eeprom):
        if not self.ADDR_BIT_RATE_NOMINAL:
            return None

        data = encode(eeprom[self.ADDR_BIT_RATE_NOMINAL])
        return struct.unpack('B', data)[0] * self.BIT_RATE_MULTIPLIER

    @classmethod
    def decode_rate_sel(self, eeprom):
        if not self.ADDR_RATE_SEL:
            return None

        data = eeprom[self.ADDR_RATE_SEL]
        rs = (struct.unpack('B', data)[0] & self.RATE_SEL_MASK) >> self.RATE_SEL_OFFSET
        return self.RATE_SEL_MAP.get(rs)

    @classmethod
    def encode_rate_sel(self, eeprom, value):
        if not self.ADDR_RATE_SEL:
            return

        if value not in self.RATE_SEL_RMAP:
            return

        # Check if rate select is implemented
        data = (struct.unpack('B', eeprom[self.ADDR_OPTION_VALUES]))[0]
        if (data & self.OPTION_RATE_SEL_MASK) == 0:
            return

        # Check if rate identifier supports SFF-8431(speed "< 4G" or " > 4G")
        data = (struct.unpack('B', eeprom[self.ADDR_RATE_IDENT]))[0]
        if (data & self.RATE_IDENT_SFF_8431_VALUE) != self.RATE_IDENT_SFF_8431_VALUE:
            return

        # Check what rate is selected.
        data = (struct.unpack('B', eeprom[self.ADDR_RATE_SEL]))[0]
        curr_rs = (data & self.RATE_SEL_MASK) >> self.RATE_SEL_OFFSET
        rs = self.RATE_SEL_RMAP.get(value)
        if curr_rs == rs:
            return

        data &= ~self.RATE_SEL_MASK
        data |= rs << self.RATE_SEL_OFFSET
        eeprom[self.ADDR_RATE_SEL] = struct.pack('B', data)
        logger.debug("speed: %d, RS 0x%x" %(value, data))


    @classmethod
    def decode_passive_or_om4_length(self, eeprom):
        data = eeprom[self.ADDR_ASSY_LENGTH]
        if len(data) != 1:
            return None

        return struct.unpack('B', data)[0]

    @classmethod
    def decode_vendor_oui(self, eeprom):
        oui = encode(eeprom[self.ADDR_OUI:self.ADDR_OUI + self.LEN_OUI])
        if len(oui) != self.LEN_OUI:
            log_throttle('failed reading vendor OUI')
            return None
        return '%02x:%02x:%02x' % struct.unpack('BBB', oui)

    @classmethod
    def decode_vendor_pn(self, eeprom):
        return eeprom[self.ADDR_PN:self.ADDR_PN + self.LEN_PN].strip()

    @classmethod
    def decode_vendor_name(self, eeprom):
        return eeprom[self.ADDR_NAME:self.ADDR_NAME + self.LEN_NAME].strip()

    @classmethod
    def decode_vendor_rev(self, eeprom):
        return eeprom[self.ADDR_REV:self.ADDR_REV + self.LEN_REV].strip()

    @classmethod
    def decode_vendor_sn(self, eeprom):
        return eeprom[self.ADDR_SN:self.ADDR_SN + self.LEN_SN].strip()


class SFPDecoder(SFFDecoder):
    ADDR_CONNECTOR_CODE = 2
    ADDR_NAME = 20
    LEN_NAME = 16
    ADDR_OUI = 37
    LEN_OUI = 3
    ADDR_PN = 40
    LEN_PN = 16
    ADDR_REV = 56
    LEN_REV = 4
    ADDR_SN = 68
    LEN_SN = 16

    ADDR_ETH10G_CODE = 3
    ETH10G_CODE_MAP = { 4 : '10g-sr' , # 10GBase-SR (multi-mode fiber)
                        5 : '10g-lr' , # 10GBase-LR (single-mode fiber)
                        6 : '10g-lrm', # 10GBase-LRM (long reach multi-mode fiber)
                        7 : '10g-er' } # 10GBase-ER (extended reach single-mode fiber)
    ADDR_ETH1G_CODE = 6
    ETH1G_CODE_MAP =  { 0 : '1000base-sx', # 1GBase-SX (multi-mode fiber)
                        1 : '1000base-lx', # 1GBase-LX (single-mode fiber)
                        2 : '1000base-cx', # 1GBase-CX (balanced copper)
                        3 : '1000base-t' } # 1GBase-T (twisted pair)

    ADDR_ETH_EXTENDED_CODE = 36  # ethernet extended compliance code (8-bit, enum)
    ADDR_PASSIVE_CABLE = 8  # bits 2 and 3 indicate a passive cable
    PASSIVE_CABLE = (3 << 2)

    ETH_CODE_DEFAULT = '10g-cr'
    ADDR_POWER_CTL = None

    ADDR_BIT_RATE_NOMINAL = 12
    BIT_RATE_MULTIPLIER = 100
    SFPP_BIT_RATE = 10300
    SFP28_BIT_RATE = 25500

    CODE_1G_DEFAULT = ETH1G_CODE_MAP[2]

    ADDR_RATE_SEL = 366
    RATE_SEL_OFFSET = 3
    RATE_SEL_MASK = 1 << RATE_SEL_OFFSET
    # Map of RS register value and speed
    RATE_SEL_MAP = {  0 : 1000,
                      1 : 10000 }
    RATE_SEL_RMAP = { v: k for k, v in RATE_SEL_MAP.items() }

    ADDR_OPTION_VALUES = 65
    OPTION_RATE_SEL_OFFSET = 5
    OPTION_RATE_SEL_MASK = 1 << OPTION_RATE_SEL_OFFSET

    ADDR_RATE_IDENT = 13
    RATE_IDENT_SFF_8431_VALUE = 0x2

    ADDR_CDR_CTRL = None
    ADDR_EXT_IDENT = None
    ADDR_CDR_CTRL_IMPLMTD = None
    # TX_Disable SW setting not implemented, use HW pin on SFP
    ADDR_TX_DISABLE = None
    ADDR_TX_DISABLE_IMPLMTD = None

    ADDR_ASSY_LENGTH = 18  # for fixed assmbly, length in 1m, for optical the supported distance over OM4 in 10m's

    TX_PWR = 0x100 + 102
    RX_PWR = 0x100 + 104

    DIAGNOSTIC_TYPE = 92
    DOM_SUPPORT = 0b01000000

    @classmethod
    def _check_dom_support(self, eeprom):
        return self.DOM_SUPPORT & ord(eeprom[self.DIAGNOSTIC_TYPE])

    @classmethod
    def decode_dom(self, eeprom):

        if not self._check_dom_support(eeprom): return {}

        def cmw(slc):
            return (ord(slc[0])*(2**8) + ord(slc[1])) / 10000.0

        tx_pwr = cmw(eeprom[self.TX_PWR:self.TX_PWR + 2])

        rx_pwr = cmw(eeprom[self.RX_PWR:self.RX_PWR + 2])

        return {"TX":(tx_pwr,), "RX":(rx_pwr,)}

    @classmethod
    def enable_high_power(self, eeprom, label):
        pass

    @classmethod
    def _decode_ethcodes(self, eeprom, addr, ethcode_map):
        data = encode(eeprom[addr])
        if len(data) != 1:
            return None

        ethcodes = struct.unpack('B', data)[0]

        return [ ethcode_map[x] for x in ethcode_map.keys()
                 if (1 << x) & ethcodes ]

    @classmethod
    def decode_ethcodes(self, eeprom):
        ethcodes = tuple(sorted(
            self._decode_ethcodes(eeprom, self.ADDR_ETH1G_CODE, self.ETH1G_CODE_MAP) +
            self._decode_ethcodes(eeprom, self.ADDR_ETH10G_CODE, self.ETH10G_CODE_MAP)))

        data = encode(eeprom[self.ADDR_PASSIVE_CABLE])
        if len(data) == 1:
            passive_code = struct.unpack('B', data)[0]
            if passive_code & self.PASSIVE_CABLE == 0:
                # Not a passive cable, extended codes are valid if any
                data = encode(eeprom[self.ADDR_ETH_EXTENDED_CODE])
                if len(data) == 1:
                    extethcodes = struct.unpack('B', data)[0]
                    if extethcodes != 0:
                        if extethcodes not in self.ETH_EXTENDED_CODE_MAP:
                            raise InvalidTransceiverCode("unrecognized extended ethcode, code: 0x%x" % extethcodes)
                        else:
                            ethcodes = tuple(sorted(ethcodes + (self.ETH_EXTENDED_CODE_MAP[extethcodes], )))
            else:
                logger.debug('%s: Passive cable' % (self.decode_vendor_pn(eeprom)))

        logger.debug('%s: ethcodes: %s' % (self.decode_vendor_pn(eeprom), ethcodes))

        # It's quite normal to assume copper as a fallback ethcode.
        # Return either 25g or 10g baseCR depending on the advertised baud
        # rate.
        if len(ethcodes) == 0 or ethcodes[0] == self.CODE_1G_DEFAULT:
            br = self.decode_bit_rate(eeprom)
            if br == self.SFPP_BIT_RATE:
                return ('10g-cr',)
            elif br == self.SFP28_BIT_RATE:
                return ('25g-cr',)
            else:
                return (self.ETH_CODE_DEFAULT, )

        return ethcodes

class QSFPDecoder(SFFDecoder):
    ADDR_CONNECTOR_CODE = 130 # external connector (8-bit enum)
    ADDR_NAME = 148
    LEN_NAME = 16
    ADDR_OUI = 165
    LEN_OUI = 3
    ADDR_PN = 168
    LEN_PN = 16
    ADDR_REV = 184
    LEN_REV = 2
    ADDR_SN = 196
    LEN_SN = 16

    # Compliance codes are at address 131 as an 8-bit bitmap. If bit 7 is set,
    # we need to look at extended compliance codes which has enum values.
    # A cable could have more than one compliance code,
    # e.g., 100G QSFP28 DAC has 2 compliance codes, 40G CR4 and 100G CR4.

    ADDR_ETH_CODE = 131 # ethernet compliance code (8-bit, mask)
    ETH_CODE_MAP = { 0 : '40g-active',     # 40G active (xlppi)
                     1 : '40g-lr4',        # 40G LR4
                     2 : '40g-sr4',        # 40G SR4
                     3 : '40g-cr4',        # 40G CR4
                     4 : '10g-sr',         # 10G SR
                     5 : '10g-lr',         # 10G LR
                     6 : '10g-lrm' }       # 10G LRM

    ETH_CODE_EXTENDED = (1 << 7)
    ADDR_ETH_EXTENDED_CODE = 192  # ethernet extended compliance code (8-bit, enum)

    ADDR_ASSY_LENGTH = 146  # cable assembly (copper or active) length, units of meters
    ETH_CODE_DEFAULT = '40g-cr4'
    ETH_EXTENDED_CODE_DEFAULT = '100g-cr4'

    # Some buggy cables (CM-5102) return all zeros or all ones.  Lets try to do
    # something predictible and assume a default.
    ETH_CODE_QUIRKS = { 0    : ETH_CODE_DEFAULT,
                        0xff : ETH_CODE_DEFAULT }
    ETH_CODE_SANCTIONED_HACKS = { }

    ETH_EXTENDED_CODE_QUIRKS = { 0    : ETH_EXTENDED_CODE_DEFAULT,}
    ETH_EXTENDED_CODE_SANCTIONED_HACKS = { }

    ADDR_POWER_CTL = 93
    POWER_CTL_LPOVERRIDE_MASK = 0x3
    POWER_CTL_LPOVERRIDE_MAP = { 0 : 'high',
                                 1 : 'override-high',
                                 2 : 'low',
                                 3 : 'override-low' }
    POWER_CTL_LPOVERRIDE_RMAP = { v: k for k, v in POWER_CTL_LPOVERRIDE_MAP.items() }
    ADDR_RATE_SEL = None

    ADDR_BIT_RATE_NOMINAL = 140
    BIT_RATE_MULTIPLIER = 100

    ADDR_CDR_CTRL = 98
    ADDR_EXT_IDENT = 129
    CDR_EXT_PRESENT_MASK = 0x0c
    ADDR_CDR_CTRL_IMPLMTD = 194
    CDR_CTRL_IMPLMTD_MASK = 0xc0

    ADDR_TX_DISABLE = 86
    ADDR_TX_DISABLE_IMPLMTD = 195
    TX_DISABLE_IMPLMTD_MASK = 0x08

    TX_DISABLE_MAP = {
        'off': 0x0f,
        'on': 0}

    # Taken from ethtool. Listed in SFF-8436 as "Reserved channel monitor set 3"
    TX1_PWR = 0x32
    TX2_PWR = 0x34
    TX3_PWR = 0x36
    TX4_PWR = 0x38

    RX1_PWR = 0x22
    RX2_PWR = 0x24
    RX3_PWR = 0x26
    RX4_PWR = 0x28

    # Map of CDR control value to desired state.  The CDR is an onboard circuit on some
    # optical modules.  It is designed to run at the cable speed, so if you run it at a lower
    # speed, it needs to be disabled (bypassed).  Thus far we only see it on 100G optical.
    # When bypassed or not implemented, the SerDes handles the CDR function.
    # RX and TX CDRs are independently controllable per lane.  We set them all on or off simultaneously.
    CDR_SEL_MAP = { 0    : 'off',
                    0xff : 'on' }
    CDR_SEL_RMAP = {v: k for k, v in CDR_SEL_MAP.items()}

    POWER_CLASS_MASK = 0xC3
    POWER_CLASS_1 = 0x00
    POWER_CLASS_2 = 0x40
    POWER_CLASS_3 = 0x80
    POWER_CLASS_4 = 0xC0
    POWER_CLASS_5 = 0xC1
    POWER_CLASS_6 = 0xC2
    POWER_CLASS_7 = 0xC3
    HIGH_POWER_CLASS_ENABLE = 0x4
    @classmethod
    def enable_high_power(self, eeprom, label):
        data = (struct.unpack('B', eeprom[self.ADDR_EXT_IDENT]))[0]
        power_class = (data & self.POWER_CLASS_MASK)
        if (power_class == self.POWER_CLASS_5 or
            power_class == self.POWER_CLASS_6 or
            power_class == self.POWER_CLASS_7):
            current = (struct.unpack('B', eeprom[self.ADDR_POWER_CTL]))[0]
            byte93 = current | self.HIGH_POWER_CLASS_ENABLE
            if byte93 != current: # only update the eeprom if the value is different
                logger.info ("Port %s set to highpower mode %x" % (label, power_class))
                eeprom[self.ADDR_POWER_CTL] = struct.pack('B', byte93)

    @classmethod
    def decode_ethcodes_base(self, ethcodes, hacks_dict, quirks_dict, eth_map_dict, ext):
        if ethcodes in hacks_dict.keys():
            hack = hacks_dict[ethcodes]
            return (hack,)

        if ethcodes in quirks_dict.keys():
            quirk = quirks_dict[ethcodes]
            log_throttle('enabling ethcode quirk: 0x%02x -> %s' % (ethcodes, quirk), level='debug')
            return (quirk,)

        if ext == 0:
            return tuple(sorted([ eth_map_dict[x] for x in range(7)
                                  if (1 << x) & ethcodes ]))
        elif ext == 1:
            try:
                return tuple([eth_map_dict[ethcodes]], )
            except KeyError:
                raise InvalidTransceiverCode("unrecognized extended ethcode, code: 0x%x" % ethcodes)

        return tuple()

    @classmethod
    def decode_dom(self, eeprom):

        #Decodes values in mW

        def cmw(slc):
            return (ord(slc[0])*(2**8) + ord(slc[1])) / 10000.0

        tx1_pwr = cmw(eeprom[self.TX1_PWR:self.TX1_PWR + 2])
        tx2_pwr = cmw(eeprom[self.TX2_PWR:self.TX2_PWR + 2])
        tx3_pwr = cmw(eeprom[self.TX3_PWR:self.TX3_PWR + 2])
        tx4_pwr = cmw(eeprom[self.TX4_PWR:self.TX4_PWR + 2])

        rx1_pwr = cmw(eeprom[self.RX1_PWR:self.RX1_PWR + 2])
        rx2_pwr = cmw(eeprom[self.RX2_PWR:self.RX2_PWR + 2])
        rx3_pwr = cmw(eeprom[self.RX3_PWR:self.RX3_PWR + 2])
        rx4_pwr = cmw(eeprom[self.RX4_PWR:self.RX4_PWR + 2])

        return {"TX": (tx1_pwr, tx2_pwr, tx3_pwr, tx4_pwr),
                "RX": (rx1_pwr, rx2_pwr, rx3_pwr, rx4_pwr)}

    @classmethod
    def decode_ethcodes(self, eeprom):
        data = encode(eeprom[self.ADDR_ETH_CODE])
        if len(data) != 1:
            return None

        ethcodes = struct.unpack('B', data)[0]

        base = self.decode_ethcodes_base(ethcodes, self.ETH_CODE_SANCTIONED_HACKS,
                                         self.ETH_CODE_QUIRKS, self.ETH_CODE_MAP, 0)
        extended = ()
        if (ethcodes & self.ETH_CODE_EXTENDED):
            data = encode(eeprom[self.ADDR_ETH_EXTENDED_CODE])
            if len(data) == 1:
                ethcodes = struct.unpack('B', data)[0]
                extended = self.decode_ethcodes_base(ethcodes, self.ETH_EXTENDED_CODE_SANCTIONED_HACKS,
                                        self.ETH_EXTENDED_CODE_QUIRKS, self.ETH_EXTENDED_CODE_MAP, 1)
        return (extended + base)

class SFFCable(object):
    ADDR_IDENTIFIER = 0x0
    IDENTIFIER_MAP = { 0x03 : SFPDecoder,
                       0x0d : QSFPDecoder,
                       0x11 : QSFPDecoder }

    IDENTIFIER_NAMES = { 0x03 : 'SFP+',
                         0x0d : 'QSFP+',
                         0x11 : 'QSFP28' }

    def __init__(self, eeprom, name=None):
        self.eeprom = eeprom
        self._ethcode_override_map = {}
        self.name = name

    def enable_high_power(self):
        decoder = self.decoder
        if not decoder:
            return

        self.decoder.enable_high_power(self.eeprom, self.name)

    def override_ethcodes(self, override_map):
        self._ethcode_override_map.update(override_map)

    def clear_eeprom_data_cache(self):
        self.eeprom.clear_data_cache()

    @property
    def dom(self):
        decoder = self.decoder
        if not decoder:
            return None

        return self.decoder.decode_dom(self.eeprom)

    @property
    def check(self):
        try:
            test = self.eeprom[self.ADDR_IDENTIFIER]
            return True
        except Exception:
            return False
        return False

    @property
    def identifier(self):
        val = struct.unpack('B', self.eeprom[self.ADDR_IDENTIFIER])[0]
        return self.IDENTIFIER_NAMES.get(val, 'Unknown')

    @property
    def decoder(self):
        data = self.eeprom[self.ADDR_IDENTIFIER]
        if len(data) != 1:
            logger.error('Port %s, failed to get identifier, got %u bytes' %
                         (self.name, len(data)))
            return None

        identifier = struct.unpack('B', encode(data))[0]
        if identifier not in self.IDENTIFIER_MAP.keys():
            log_throttle('Port %s, invalid SFF identifier: 0x%02x' %
                         (self.name, identifier))
            return None

        return self.IDENTIFIER_MAP[identifier]

    @property
    def ouipn(self):
        decoder = self.decoder
        if not decoder:
            return

        return '%s,%s' % (decoder.decode_vendor_oui(self.eeprom),
                          decoder.decode_vendor_pn(self.eeprom))

    @property
    def vendorpn(self):
        decoder = self.decoder
        if not decoder:
            return

        return '%s,%s' % (decoder.decode_vendor_name(self.eeprom),
                          decoder.decode_vendor_pn(self.eeprom))

    @property
    def ethcodes(self):
        if self._ethcode_override_map is not None:
            if 'port' in self._ethcode_override_map:
                override = self._ethcode_override_map['port']
                log_throttle('Port %s, enabling port ethcode override: %s' %
                             (self.name, override[0]), once=True, level='info')
                return override
            elif self.ouipn in self._ethcode_override_map:
                override = self._ethcode_override_map[self.ouipn]
                log_throttle('Port %s, enabling ouipn ethcode override: %s -> %s' %
                             (self.name, self.ouipn, override[0]), once=True, level='info')
                return override
            elif self.vendorpn in self._ethcode_override_map:
                override = self._ethcode_override_map[self.vendorpn]
                log_throttle('Port %s, enabling vendorpn ethcode override: %s -> %s' %
                             (self.name, self.vendorpn, override[0]), once=True, level='info')
                return override

        decoder = self.decoder
        if not decoder:
            return tuple()

        return decoder.decode_ethcodes(self.eeprom)

    @property
    def ethcodes_safe(self):
        try:
            return self.ethcodes
        except (EEPROMTimeout, InvalidTransceiverCode):
            return tuple()

    @property
    def connector(self):
        decoder = self.decoder
        if not decoder:
            return tuple()

        return decoder.decode_connector(self.eeprom)

    def _get_lpoverride(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_lpoverride(self.eeprom)

    def _set_lpoverride(self, value):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.encode_lpoverride(self.eeprom, value)

    lpoverride = property(_get_lpoverride, _set_lpoverride)

    def _get_rate_sel(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_rate_sel(self.eeprom)

    def _set_rate_sel(self, value):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.encode_rate_sel(self.eeprom, value)

    rate_sel = property(_get_rate_sel, _set_rate_sel)

    def cdr_present(self):
        decoder = self.decoder
        if not decoder:
            return None
        return decoder.decode_cdr_present(self.eeprom)

    def cdr_controllable(self):
        decoder = self.decoder
        if not decoder:
            return None
        return decoder.decode_cdr_controllable(self.eeprom)

    def _get_cdr_value(self):
        decoder = self.decoder
        if not decoder:
            return None
        return decoder.decode_cdr_setting(self.eeprom)

    def _set_cdr_value(self, value):
        decoder = self.decoder
        if not decoder:
            return None

        curr_cdr = decoder.decode_cdr_setting(self.eeprom)
        if curr_cdr != value:
            logger.debug('Changing CDR value from %s to %s' % (curr_cdr, value))
            decoder.encode_cdr_setting(self.eeprom, value)

    cdr_value = property(_get_cdr_value, _set_cdr_value)

    @property
    def toggle_tx_disable(self):
        decoder = self.decoder
        if not decoder:
            return None

        if decoder.decode_tx_disable_implmtd(self.eeprom):
            decoder.encode_tx_disable(self.eeprom, 'off')
            time.sleep(1)
            decoder.encode_tx_disable(self.eeprom, 'on')
        else:
            logger.debug('module does not have tx_disable implemented')

    @property
    def cable_assy_length(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_passive_or_om4_length(self.eeprom)

    @property
    def vendor_oui(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_vendor_oui(self.eeprom)

    @property
    def vendor_pn(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_vendor_pn(self.eeprom)

    @property
    def vendor_name(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_vendor_name(self.eeprom)

    @property
    def vendor_rev(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_vendor_rev(self.eeprom)

    @property
    def vendor_sn(self):
        decoder = self.decoder
        if not decoder:
            return None

        return decoder.decode_vendor_sn(self.eeprom)


class SysFSEEPROM(object):
    def __init__(self, path):
        self.path = path
        self.eeprom_cached_data = {}

    def clear_data_cache(self):
        self.eeprom_cached_data = {}

    def _getslice(self, sl):
        # eeproms are slow, only read exactly the number of bytes requested.
        # (no buffering)
        read_len = sl.stop - sl.start
        if (sl.start in self.eeprom_cached_data) and (len(self.eeprom_cached_data[sl.start]) == read_len): # if the base addr is in the cache  and the length matches exactly
            return self.eeprom_cached_data[sl.start]             # return this request from cache
        else:                                                  # else, read it out of eeprom and fill the cache
            try:
                eeprom_rd_fd = io.open(self.path, 'rb', buffering = 0)
                eeprom_rd_fd.seek(sl.start)
                rv = eeprom_rd_fd.read(read_len)
                self.eeprom_cached_data[sl.start] = rv;
                return rv
            except IOError as e:
                if e.errno == errno.ETIMEDOUT:
                    raise EEPROMTimeout('timeout reading eeprom at: %s' % self.path)
                    return ''
                else:
                    raise e

    def _setslice(self, sl, value):
        try:
            eeprom_wr_fd = io.open(self.path, 'wb', buffering = 0)
            eeprom_wr_fd.seek(sl.start)
            rv = eeprom_wr_fd.write(value)
            # write-through cache, all writes auto clean the line
            self.eeprom_cached_data[sl.start] = value
            return rv
        except IOError as e:
            if e.errno == errno.ETIMEDOUT:
                raise EEPROMTimeout('timeout writing eeprom at: %s' % self.path)
            else:
                raise e

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self._getslice(key)
        else:
            return self._getslice(slice(key, key + 1))

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            self._setslice(key, value)
        else:
            return self._setslice(slice(key, key + 1), value)

    @property
    def bus(self):
        base = os.path.split(self.path)[0]
        devpath = os.path.join(base, '..', 'i2c-dev')
        for directory in os.listdir(devpath):
            if directory.startswith('i2c-'):
                return int(directory[4:])

        raise NoI2CBus('no i2c bus path for: %s' % self.path)
