#!/usr/bin/env python

# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import json
import logging
import logging.config
import re
import sys
import time
import subprocess
import socket
import os.path
from os import path
from pprint import pformat

# Safely imports modules whose names changed from Python 2 -> 3
try:
    import configparser as ConfigParser
except ImportError:
    import ConfigParser

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from cl_parsers.parse import parse
from l1_show.eeprom_utils import SFFCable

from module_qualify.qualify_utils import swp_sort_pattern


this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error

compliance_conversion = {
    '100g-sr4': ['25g-sr'],
    '100g-lr4': ['25g-lr'],
    '100g-er4': ['25g-er'],
    '100g-active-cu': ['25g-active-cu'],
    '100g-cr4': ['25g-cr', '50g-cr2'],
    '100g-active': ['25g-active']
}


def Dut(cl_support):
    # Does imports here to avoid import errors, since `cumulus` module may not
    # exist on non-live systems running this command
    if cl_support is None:
        from .live import Live as Switch
        return Switch()
    else:
        from .cl_support import Cl_Support as Switch
        return Switch(cl_support)


def update_if_none(dct, key, val):
    if (key not in dct) or (dct[key] is None):
        dct[key] = val


def critical(mesg):
    this_log.critical(mesg)
    this_log.critical('Exiting abnormally')
    sys.exit(1)


def debug_test():
    debug('Logging testing of dut_obj submodule')


def setup_logging(config_file_path=None,
                  log_file_dir='./',
                  quiet=False,
                  default_level=logging.INFO,
                  logformat='%(asctime)s: %(name)20s: %(levelname)s: %(message)s'):
    """
    Setup logging config for a module using a JSON config file specified by config_file_path
      and a log directory target specified by log_file_dir

    The default configuration included in this module configures the following loggers:
      autolib.netobjects for DUT output:  outputs to 'autolib.log' file with level DEBUG
      main module (initial called script): outputs to 'info.log' file with level INFO
      paramiko.transport: paramiko outputs are set to ERROR level to ignore most messages and outputs to 'info.log' file
    a few other handlers are defined for further customization.

    :param config_file_path: the location of a logging config file to load.  Default is the one included in this repo.
    :param log_file_dir: the directory to write the files to.  Default is the present working directory
    :param quiet: if set to True, suppress the initial setup messages describing the location of the config and log files
    :param default_level: In absence of a valid config_file_path file, this sets a basic logger of this level
    :param logformat: The default log message format to use
    
    return the config_dct in case the user wants to reconfigure
    """
    if not config_file_path:
        config_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'default_log_config.json')
    if not quiet:
        print('log_utils: Reading logging config file from: {}'.format(config_file_path))
    config_file_output = read_local_file(config_file_path)
    if config_file_output:
        config = json.loads(config_file_output)
        config_dct = _insert_logging_path(config, log_file_dir, quiet)
        logging.config.dictConfig(config_dct)
        return config_dct
    else:
        if not quiet:
            print('log_utils: config_file_path did not exist, setting basicConfig, level={}'.format(default_level))

        logging.basicConfig(level=default_level, format=logformat)


def _insert_logging_path(config_dct, log_file_dir, quiet=False):
    """
    Internal function to change the logging config to send the logs to a specified directory
    :param config_dct: Logging Dictionary to modify with log_file_dir
    :param log_file_dir: Logging dir path to insert in log config in order to control log locations
    :param quiet: Set to True to suppress pre-log printed messages
    :return:
    """
    for handler_name in config_dct['handlers']:
        for key in config_dct['handlers'][handler_name]:
            if key == 'filename':
                old_filename = config_dct['handlers'][handler_name][key]
                new_filename = os.path.join(log_file_dir, old_filename)
                config_dct['handlers'][handler_name][key] = new_filename
                if not quiet:
                    print('log_utils: Setting logging path for handler {} to: {}'.format(handler_name, new_filename))

    return config_dct


def debug_logging(modules, this_log=None):
    if not isinstance(modules, list):
        modules = [modules]

    if not this_log:
        this_log = logging.getLogger()

    if 'all' in modules:
        for name in sorted(logging.Logger.manager.loggerDict):
            logger_to_change = logging.getLogger(name)
            this_log.info('Logger {} Level: {}. Changing to DEBUG'.format(
                name, logging.getLevelName(logger_to_change.getEffectiveLevel())))
            logger_to_change.setLevel('DEBUG')

    else:
        for module in modules:
            if 'main' in module:
                module = '__main__'
            logger_to_change = logging.getLogger(module)
            this_log.info('Logger {} Level: {}. Changing to DEBUG'.format(
                module, logging.getLevelName(logger_to_change.getEffectiveLevel())))
            logger_to_change.setLevel('DEBUG')
            if 'main' in module:
                this_log.debug('Doc info:')
                this_log.debug(__doc__)


def list_logger_levels(logger_list=logging.Logger.manager.loggerDict):
    if not isinstance(logger_list, list) and not isinstance(logger_list, dict):
        logger_list = [logger_list]

    for logger_to_read in logger_list:
        info('Logger: {} Level: {}'.format(logger_to_read,
                                           logging.getLevelName(logging.getLogger(logger_to_read).getEffectiveLevel())
                                           ))


def format_cmd(cmd):
    pattern = re.compile(r''' [01]: ''')
    return pattern.sub(' ', cmd)


def format_output(output):
    return output.strip()


def shell_out_catch_errors(input_line):
    try:
        output = format_output(subprocess.check_output(input_line, shell=True))

    except subprocess.CalledProcessError as e:
        # error executing, collect any output there was
        output = e.output

    return output.decode()


def parse_to_dct(dut, cmd, *args, **kwargs):
    out = shell_out_catch_errors(cmd)
    cmd = format_cmd(cmd)
    return parse(cmd, out)
    # TODO substitute in below code to prevent crash on bad parses
    # parsed_out = parse(cmd, out)
    # if parsed_out is None:
    #     return {}
    # else:
    #     return parsed_out


def read_local_file(filename, mode='rt'):
    try:
        with open(filename, mode) as f:
            return f.read().strip()
    except IOError:
        debug('File: {} not found'.format(filename))
        return None


def find_path(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return root


def dut_obj_install_string(dut, stringname, string_to_attach):
    if not hasattr(dut, stringname):
        setattr(dut, stringname, str())

    debug('Adding string: \"{0}\" to object: \"{1}\"'.format(stringname, dut.hostname))
    setattr(dut, stringname, string_to_attach)


def dut_obj_install_list(dut, listname, list_to_attach):
    if not hasattr(dut, listname):
        setattr(dut, listname, list())

    debug('Adding list \"{0}\" to \"{1}\" object'.format(listname, dut.hostname))
    setattr(dut, listname, list_to_attach)


def dut_obj_install_dict(dut, dictname, dict_to_attach):
    if not hasattr(dut, dictname):
        setattr(dut, dictname, dict())

    debug('Adding dict \"{0}\" to \"{1}\" object'.format(dictname, dut.hostname))
    setattr(dut, dictname, dict_to_attach)


def read_portwd_config(configfile):
    if not path.exists(configfile):
        return None, None

    config = ConfigParser.SafeConfigParser()
    config.optionxform = str
    configbuf = read_local_file(configfile)
    if not configbuf:
        return None, None
    configfp = StringIO(configbuf.replace(':', '%%3A'))
    config.readfp(configfp)

    try:
        cable_overrides = dict(config.items('cables'))
    except ConfigParser.NoSectionError:
        cable_overrides = {}
    try:
        port_overrides = dict(config.items('ports'))
    except ConfigParser.NoSectionError:
        port_overrides = {}

    if len(cable_overrides):
        cable_overrides = {k.replace('%%3A', ':'): (v,) for (k, v) in cable_overrides.items()}

    debug('Cable overrides: {}'.format(cable_overrides))
    debug('Port overrides: {}'.format(port_overrides))
    return cable_overrides, port_overrides


def setup_overrides(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        cable_override_dct = {}
        port_override_dct = {}

        configfiles = config_file_list(dut)
        for configfile in configfiles:
            cable_override, port_override = read_portwd_config(configfile)
            if cable_override:
                cable_override_dct.update(cable_override)
            if port_override:
                port_override_dct.update(port_override)

        dut_obj_install_dict(dut, 'cable_override', cable_override_dct)
        dut_obj_install_dict(dut, 'port_override', port_override_dct)


def config_file_list(dut):
    # on switch:  '/usr/share/cumulus/portwd.conf', '/etc/cumulus/portwd.conf'
    configfiles = ['/usr/share/cumulus/portwd.conf']
    # on laptop:  local portwd.conf,  'cl_support:/etc/cumulus/portwd.conf'
    portwd_conf_path = find_in_syspath('portwd/portwd.conf')
    if portwd_conf_path:
        configfiles = [portwd_conf_path]

    # if there is a customer etc/cumulus/portwd.conf, add it
    local_conf_file = dut.get_local_portwd_conf()
    if local_conf_file:
        configfiles.append(local_conf_file)

    return configfiles


def find_in_syspath(filename):
    for i in sys.path:
        thepath = os.path.join(i, filename)
        if os.path.exists(thepath):
            return thepath
    return False


def convert_parsed_porttab(parsed_porttab_dct, bcm_show_portmap_dct):
    """
    The porttab parser should return a complex dictionary with multiple fields, but we only
    need the mapping swpXX <-> unitZ xeYY.  However, Tomahawk and Trident2+ follow a new set of rules
    So handle both varieties here.
    """
    phy_dct = {}
    for unit in bcm_show_portmap_dct:
        for phy in bcm_show_portmap_dct[unit]:
            logical_port = bcm_show_portmap_dct[unit][phy]['logical']
            phy_dct[logical_port] = (phy, unit)

    porttab_dct = {}
    for port in parsed_porttab_dct:
        sdk_intf = parsed_porttab_dct[port]['sdk_intf']
        logical_port = parsed_porttab_dct[port]['logical_port']
        unit = 'unit' + parsed_porttab_dct[port]['unit']
        if logical_port == '-1' or logical_port == '0' or not logical_port:
            phy = sdk_intf
        else:
            phy, unit = phy_dct[logical_port]

        porttab_dct[port] = (phy, unit)

    return porttab_dct


def dut_install_porttab(dut_devs=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        porttab_dct = dut.get_porttab()
        debug('Installing porttab_dct to {} with keys {}'.format(
            dut.hostname, porttab_dct.keys()))
        if not porttab_dct:
            critical('porttab_dct is Null')
        dut_obj_install_dict(dut, 'porttab_dct', porttab_dct)


def dut_install_eeprom_list(dut_devs=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        eeprom_list = dut.get_eeprom_list()
        eeprom_list = sorted(list(set(eeprom_list)), key=lambda l: int(l))

        debug('Adding eeprom_list: {} to DUT obj: {}'.format(eeprom_list, dut.hostname))
        dut_obj_install_list(dut, 'eeprom_list', eeprom_list)


def dut_install_ethtool_m(dut_devs=None, port_list=None, write_out=False):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in dut_devs:
        if not hasattr(dut, 'eeprom_list'):
            dut_install_eeprom_list(dut)
        debug('Adding \"ethtool -m\" for each port to DUT obj: {}'.format(dut.hostname))
        ethtool_m_dct = {}
        if not port_list:
            port_list = sorted(dut.porttab_dct, key=lambda l: swp_sort_pattern(l))
        for port in port_list:
            # skip ethtool -m for fixed ports.
            if (dut.platform_object.name == 'accton,minipack' or
                    str(port).replace('swp', '').split('s')[0] in dut.eeprom_list):
                ethtool_m_dct[port] = dut.ethtool(port, '-m')

        dut_obj_install_dict(dut, 'ethtool_m_dct', ethtool_m_dct)


def dut_install_ethtool(dut_devs=None, port_list=None, write_out=False):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in dut_devs:
        debug('Adding \"ethtool\" for each port to DUT obj: {}'.format(dut.hostname))
        ethtool_dct = {}
        if not port_list:
            port_list = sorted(dut.porttab_dct, key=lambda l: swp_sort_pattern(l))
        for port in port_list:
            ethtool_dct[port] = dut.ethtool(port)

        dut_obj_install_dict(dut, 'ethtool_dct', ethtool_dct)


def dut_install_os_release(dut_devs=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        debug('Reading /etc/os-release on DUT: {}'.format(dut.hostname))
        filename = os.path.join(dut.path_prefix, 'etc/os-release')
        os_release = read_local_file(filename)

        # First fallback file - /etc/os-release is a symlink that isn't collected
        # in cl-support, so this is normal
        if os_release is None:
            debug('/etc/os-release file is not found')
            filename = os.path.join(dut.path_prefix, 'etc/cumulus/etc.replace/os-release')
            os_release = read_local_file(filename)

        # Second fallback file - if the other files can't be found, we can infer
        # some info from the lsb_release. Currently, all we need is the version
        if os_release is None:
            warn('/etc/cumulus/etc.replace/os-release file is not found')
            filename = os.path.join(dut.path_prefix, 'Support/lsb_release')
            data = read_local_file(filename)
            if data is None:
                warn('Cannot determine os-release')
                os_release = ''
            else:
                match = re.search('^Release:\s+(\S+)$', data, re.MULTILINE)
                if match is not None:
                    os_release = 'VERSION_ID=' + match.group(1)

        debug('Adding \"os-release\" {} to DUT obj: {}'.format(os_release, dut.hostname))
        dut_obj_install_string(dut, 'os_release', os_release)


def dut_install_ip_link(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        ip_link_dct = dut.ip_link_show()
        dut_obj_install_dict(dut, 'ip_link_dct', ip_link_dct)


def translate_ip_link_status(dut, port):
    if not hasattr(dut, 'ip_link_dct'):
        dut_install_ip_link(dut)

    if port not in dut.ip_link_dct:
        return 'ERROR', 'ERROR'
    debug('ip_link flags: {}'.format(dut.ip_link_dct[port].get('flags', None)))
    if dut.ip_link_dct[port].get('flags', None) == 'BROADCAST,MULTICAST':
        return 'Admin Down', 'Down'
    if dut.ip_link_dct[port].get('state', None) == 'UP':
        return 'Admin Up', 'Up'
    if dut.ip_link_dct[port].get('state', None) == 'DOWN':
        return 'Admin Up', 'Down'
    return 'ERROR', 'ERROR'


def dut_install_ethmodes(dut_devs, port_list=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in dut_devs:
        debug('Adding \"ethmodes\" for each port {} to DUT obj: {}'.format(port_list, dut.hostname))
        ethmode_dct = {}
        if not port_list:
            port_list = sorted(dut.porttab_dct, key=lambda l: swp_sort_pattern(l))
        for port in port_list:
            fuse = os.path.join(dut.path_prefix, 'cumulus/switchd/config/interface/{port}/ethtool_mode'.format(port=port))
            output = read_local_file(fuse)
            if output:
                ethmode_dct[port] = output

            debug('Port: {}, ethmode: {}'.format(port, output))

        dut_obj_install_dict(dut, 'ethmode_dct', ethmode_dct)


def dut_install_eeprom(dut_devs=None, port_list=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in dut_devs:
        if not hasattr(dut, 'eeprom_list'):
            dut_install_eeprom_list(dut)
        debug('Adding \"eeprom hex output\" for each port {} to DUT obj: {}'.format(port_list, dut.hostname))
        eeprom_dct = {}
        if not port_list:
            port_list = sorted(dut.porttab_dct, key=lambda l: swp_sort_pattern(l))
        for port in port_list:
            # skip ethtool -m for fixed ports.
            if (dut.platform_object.name == 'accton,minipack') or (
                    dut.asic_vendor == 'mlx') or (
                    str(port).replace('swp', '').split('s')[0] in dut.eeprom_list):
                eeprom_port_dct = dut.ethtool(port, '-m hex on')
                if not eeprom_port_dct:
                    continue
                eeprom_dct[port] = eeprom_port_dct
                debug('Port: {}, EEPROM:'.format(port) + pformat(eeprom_port_dct))
        dut_obj_install_dict(dut, 'eeprom_dct', eeprom_dct)


def dut_install_ethtool_showfec(dut_devs=None, port_list=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    active_fec_pattern = re.compile('^(Active\s)?FEC\sencoding(s)?\s*:\s(?P<active_fec>\w+)$')
    config_fec_pattern = re.compile('Configured FEC encodings:\s(?P<config_fec>\w+)')

    for dut in dut_devs:
        showfec_dct = {}

        debug('Adding \"ethtool --show-fec output\" for each port {} to DUT obj: {}'.format(port_list, dut.hostname))
        if not port_list:
            port_list = sorted(dut.porttab_dct, key=lambda l: swp_sort_pattern(l))
        for port in port_list:
            showfec_dct[port] = {}

            if dut.version[0:1] > '3':
                fuse = os.path.join(dut.path_prefix, 'cumulus/switchd/config/interface/{port}/active_fec'.format(port=port))
            else:
                fuse = os.path.join(dut.path_prefix, 'cumulus/switchd/config/interface/{port}/fec'.format(port=port))
            showfec_dct[port]['hw_active_fec'] = read_local_file(fuse)
            output = dut.ethtool(port, '--show-fec', False)
            if not output:
                continue
            for line in output.split('\n'):
                active_fec_match = active_fec_pattern.match(line)
                if active_fec_match:
                    showfec_dct[port]['active_fec'] = active_fec_match.group('active_fec')
                if dut.version[0:1] > '3':
                    config_fec_match = config_fec_pattern.match(line)
                    if config_fec_match:
                        showfec_dct[port]['config_fec'] = config_fec_match.group('config_fec')

            default = dut.get_default_fec_value()
            update_if_none(showfec_dct[port], 'active_fec', default)
            update_if_none(showfec_dct[port], 'config_fec', default)
            update_if_none(showfec_dct[port], 'hw_active_fec', default)

            debug('Dut: {} Port: {}, Active FEC {}, Config FEC: Hw FEC'.format(
                dut.hostname, port,
                showfec_dct[port]['active_fec'], showfec_dct[port]['config_fec'],
                showfec_dct[port]['hw_active_fec']
            ))

        dut_obj_install_dict(dut, 'showfec_dct', showfec_dct)


def get_dom(eeprom):
    if len(eeprom) == 0:
        return {}
    cable = SFFCable(eeprom)
    return cable.dom


def convert_compliance_speed(compliance_codes, hardware_speed):
    new_compliance_codes = []
    hardware_speed = str(hardware_speed).split('G')[0]
    for code in compliance_codes:
        new_compliance_codes.append(code)
        # 25G//50G/100G share many of the same compliance codes. Add the lower speed code if speed < code
        if code in compliance_conversion:
            for new_code in compliance_conversion[code]:
                if hardware_speed in new_code:
                    debug('interface speed: {}, compliance code {}. Appending code: {} to compliance code list'.format(
                        hardware_speed, code, new_code))
                    new_compliance_codes.append(new_code)

    return new_compliance_codes


def convert_mlx_autoneg(mlx_autoneg):
    if mlx_autoneg == "ON":
        return 'On (Autodetect enabld)', 'Enabled'
    elif 'FORCE' in mlx_autoneg:
        return 'Off', mlx_autoneg
    else:
        return str(mlx_autoneg).capitalize(), str(mlx_autoneg).capitalize()


def convert_fec_str(fec):
    if not fec:
        return 'error'
    elif 'not_avail' in fec:
        return 'N/A'
    elif 'No FEC' in fec:
        return 'None'
    elif 'Firecode FEC' in fec:
        return 'BaseR/FireCode'
    elif 'BaseR' in fec:
        return 'BaseR/FireCode'
    elif 'RS-FEC' in fec:
        return 'RS'
    elif 'N/A' in fec:
        return 'None (down)'
    else:
        return fec


def convert_hw_fec_str(hw_fec):
    if hw_fec == 'CL91':
        return 'CL91 (RS)'
    if hw_fec == 'CL74':
        return 'CL74 (BaseR)'
    return hw_fec


def convert_signal_detect(signal_detect):
    try:
        if signal_detect is None or signal_detect == '' or signal_detect == 'N/A':
            return 'NA '
        elif int(signal_detect) == 1:
            return 'Y'
        elif int(signal_detect) == 0:
            return 'N'
        else:
            debug('Unexpected value for signal detect: {}'.format(signal_detect))
            return 'Err'
    except Exception as e:
        debug('Unexpected value for signal detect: {}'.format(e))
        return 'Err'


def convert_kern_speed(kern_speed):
    return kern_speed.replace('000Mb/s', 'G')


def determine_hw_speed(dut, port):
    if dut.asic_vendor == 'mlx':
        if dut.mlxlink_dct[port] is None:
            dut.mlxlink_dct[port] = {}
        hw_speed = dut.mlxlink_dct[port].get('Operational Info', {'Speed': 'N/A'}).get('Speed', None)
        hw_speed = '1G' if hw_speed == 'CX' else hw_speed
        hw_speed = hw_speed.split('bE')[0]
    elif dut.asic_vendor == 'bcm':
        bcm_port, unit = dut.porttab_dct[port]
        if not dut.bcm_portstat_dct or not dut.bcm_portstat_dct[unit]:
            hw_speed = 'N/A'
        else:
            hw_speed = dut.bcm_portstat_dct.get(unit, {}).get(bcm_port, {}).get('speed', None)
    else:
        return 'Error'

    return hw_speed


def get_compliance_codes(eeprom):
    if len(eeprom) == 0:
        return []
    cable = SFFCable(eeprom)
    codes = cable.ethcodes_safe
    return [codes[x] for x in range(0, len(codes))]


def dut_install_ptmctl(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        ptmctl_dct = dut.ptmctl()
        dut_obj_install_dict(dut, 'ptmctl_dct', ptmctl_dct)


def dut_install_lldpctl(dut_devs=None, force_refresh=False):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        refresh = False
        now = time.time()
        if force_refresh:
            refresh = True

        elif not hasattr(dut, 'lldp_dct') or not hasattr(dut, 'lldp_refresh_time'):
            refresh = True

        elif not refresh and now - dut.lldp_refresh_time > 5:
                refresh = True

        if refresh:
            debug('Adding/Refreshing lldp_dct on dut: {}'.format(dut.hostname))
            lldp_dct = dut.lldpctl()
            dut_obj_install_dict(dut, 'lldp_dct', lldp_dct)
            dut_obj_install_string(dut, 'lldp_refresh_time', now)


def lldp_keyvalue_parser(output):
    lldp_dct = {}
    port_desc = re.compile(r'''lldp\.(sw[0-9]*p[0-9s]+)\.port\.descr''')
    chassis_name = re.compile(r'''lldp\.(sw[0-9]*p[0-9s]+)\.chassis\.name''')

    for line in output.split('\n'):
        if re.search('/root/.bashrc:', line):
            debug('bashrc error -skip this line')
            continue
        if '=' not in line:
            error('lldpctl line does not contain key.value=data \n {}'.format(line))
            continue

        key, value = line.split('=', 1)
        port_match = port_desc.match(key)
        if port_match:
            local_port = port_match.group(1)
            remote_port = value
            if local_port not in lldp_dct:
                lldp_dct[local_port] = {}
            lldp_dct[local_port]['neighbor_port'] = remote_port
            debug('Local port: {}. neighbor port: {}'.format(local_port, remote_port))
            continue

        chassis_match = chassis_name.match(key)
        if chassis_match:
            local_port = chassis_match.group(1)
            remote_switch = value
            if local_port not in lldp_dct:
                lldp_dct[local_port] = {}
            lldp_dct[local_port]['neighbor_switch'] = remote_switch
            debug('Local port: {}. neighbor switch: {}'.format(local_port, remote_switch))
            continue

    return lldp_dct


def lldp_remote_port_lookup(local_dut, local_port, force_refresh=False):
    dut_install_lldpctl(local_dut, force_refresh=force_refresh)

    neighbor_switch = local_dut.lldp_dct.get(local_port, {}).get('neighbor_switch', None)
    if neighbor_switch:
        neighbor_switch = neighbor_switch.split('.')[0]
    neighbor_port = local_dut.lldp_dct.get(local_port, {}).get('neighbor_port', None)
    debug('LLDP: Dut: {} Port: {}. Neighbor is {} {}'.format(local_dut.hostname, local_port,
                                                             neighbor_switch, neighbor_port))
    return neighbor_switch, neighbor_port


def dut_install_mlx_ports_info(dut_devs=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in mlx_only(dut_devs):
        debug('Adding \"mlx ports info\" to DUT obj: {}'.format(dut.hostname))
        mlx_ports_info_dct = parse_to_dct(dut, '/usr/lib/cumulus/mlxcmd ports info')
        debug('Adding mlx_ports_info_dct to {} with keys {}'.format(
            dut.hostname, mlx_ports_info_dct.keys()))
        if not mlx_ports_info_dct:
            critical('mlx_ports_info_dct is Null')
        dut_obj_install_dict(dut, 'mlx_ports_info_dct', mlx_ports_info_dct)


def is_hurricane2(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    return [dut for dut in dut_devs if dut.asic == 'Hurricane2']


def bcm_only(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    return [dut for dut in dut_devs if dut.asic_vendor == 'bcm']


def mlx_only(dut_devs):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    return [dut for dut in dut_devs if dut.asic_vendor == 'mlx']


def mlxlink_port(port):
    mlx_port = port[3:]
    if 's' in mlx_port:
        mlx_lane = mlx_port.split('s')[1]
        if mlx_lane == '0':
            mlx_port = mlx_port[0:-2]
        else:
            mlx_port = mlx_port[0:-2] + '/' + str(int(mlx_lane) + 1)
    return mlx_port


def dut_install_mlxlink(dut_devs=None, port_list=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in mlx_only(dut_devs):
        mlxlink_dct = {}
        for port in port_list:
            mlx_port = mlxlink_port(port)
            mlxlink_dct[port] = dut.mlxlink(mlx_port)
            if not mlxlink_dct[port]:
                debug('mlxlink parser returned None for port {}'.format(port))

        dut_obj_install_dict(dut, 'mlxlink_dct', mlxlink_dct)


def dut_install_version(dut_devs=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    for dut in dut_devs:
        os_version = get_version(dut)
        debug('Adding \"VERSION_ID\" {} to DUT obj: {}'.format(os_version, dut.hostname))
        dut_obj_install_string(dut, 'version', os_version)


def get_version(dut):
    if not hasattr(dut, 'os_release'):
        dut_install_os_release(dut)

    if not dut.os_release:
        return '2.x'

    for line in dut.os_release.split('\n'):
        if 'VERSION_ID' in line:
            return line.split('=')[1]

    warn('Could not determine os-release version')
    return None


# TODO: This function appears to be unused
def dut_install_bcm_phy_diag_dsc(dut_devs=None, port_list=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]
    if not isinstance(port_list, list):
        port_list = [port_list]

    for dut in bcm_only(dut_devs):
        bcm_phy_diag_dsc_dct = {}
        if not hasattr(dut, 'porttab_dct'):
            dut_install_porttab(dut)
        if dut.version[0:1] < '3':
            continue
        if port_list == ['all']:
            port_list = warp_core_list(dut.porttab_dct)

        debug('Adding Port HW state in \"bcm phy diag X dsc\" to DUT obj: {}'.format(dut.hostname))
        for port in port_list:
            phy, unit = dut.porttab_dct[port]
            cmd = '/usr/lib/cumulus/bcmcmd {} phy diag {} dsc'.format(unit[-1:]+':', phy)
            bcm_phy_diag_dsc_dct[port] = dut.bcmcmd_phy(phy, unit[-1:]+':', cmd)
            debug('Adding bcm_phy_diag_dsc_dct to {} with keys {}'.format(
                dut.hostname, bcm_phy_diag_dsc_dct.keys()))
            if not bcm_phy_diag_dsc_dct:
                critical('bcm_phy_diag_dsc_dct is Null')

        dut_obj_install_dict(dut, 'bcm_phy_diag_dsc_dct', bcm_phy_diag_dsc_dct)


def warp_core_list(swp_list):
    """For QSFPs, return a list with only the 1st phy in the warp core"""
    sorted_swp_list = sorted(swp_list, key=lambda l: swp_sort_pattern(l))
    if sorted_swp_list[-1][3:5] not in ['32', '16']:
        out_list = []
        for port in sorted_swp_list:
            if int(port[3:5]) in range(1, 48, 4) or (
                    int(port[3:5]) > 48 and port[-2:] not in ['s1', 's2', 's3']):
                out_list.append(port)

    else:
        out_list = []
        for port in sorted_swp_list:
            if port[-2:] not in ['s1', 's2', 's3']:
                out_list.append(port)

    return out_list


def dut_install_bcm_dct(dut_devs=None, cmd=None, dut_obj_property=None):
    if not isinstance(dut_devs, list):
        dut_devs = [dut_devs]

    if not cmd:
        critical('dut_install_bcm_dct called without any command given')
        return

    file_suffix = 'bcm_' + '_'.join(cmd.split())
    if not dut_obj_property:
        dut_obj_property = '_'.join([file_suffix, 'dct'])

    for dut in bcm_only(dut_devs):
        bcm_cmd_dct = dict()
        debug('Adding Port HW state in \"{}\" to DUT obj: {}'.format(cmd, dut.hostname))
        bcm_cmd_dct['unit0'] = dut.bcmcmd(0, cmd)
        if not bcm_cmd_dct['unit0']:
            if 'show counters' in cmd:
                warn('{} is Null'.format(dut_obj_property))
            else:
                error('{} is Null'.format(dut_obj_property))
                error('DUT: {}: bcmcmd {} failed to parse. Some output data will be wrong or missing.'.format(dut.hostname, cmd))
        else:
            debug('Adding unit0 {} to {} with keys {}'.format(
                dut_obj_property, dut.hostname, bcm_cmd_dct['unit0'].keys()))

        if is_hurricane2(dut):
            bcm_cmd_dct['unit1'] = dut.bcmcmd(1, cmd)
            debug('Adding unit1 {} to {} with keys {}'.format(
                dut_obj_property, dut.hostname, bcm_cmd_dct.keys()))
            if not bcm_cmd_dct['unit1']:
                if 'show counters' in cmd:
                    warn('Unit1: {} is Null'.format(dut_obj_property))
                else:
                    critical('DUT: {}. Unit1: {} is Null'.format(dut_obj_property, dut.hostname))

        dut_obj_install_dict(dut, dut_obj_property, bcm_cmd_dct)


