# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import socket
import sys
import errno
import re
from cumulus.platforms import probe
import cumulus.porttab

from l1_show.base import Base
from l1_show.l1_show_lib import *

try:
    from python_sdk_api.sx_api import *
except ImportError:
    pass

this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error


class Live(Base):

    def __init__(self):
        Base.__init__(self)

    def get_hostname(self):
        return socket.gethostname()

    def get_default_fec_value(self):
        return None

    def get_support_dir(self):
        return None

    def get_platform_object(self):
        return probe()

    def bcmcmd(self, num, cmd):
        return parse_to_dct(self, '/usr/lib/cumulus/bcmcmd {}: {}'.format(num, cmd), quiet=True)

    def bcmcmd_phy(self, port, unit, cmd):
        output = shell_out_catch_errors(cmd)
        debug('dsc output: {}'.format(output))
        return parse(cmd, output)

    def ethtool(self, port, flag=None, parse=True):
        if flag is None:
            return parse_to_dct(self, 'ethtool {}'.format(port), quiet=True)
        elif flag == '-m':
            return parse_to_dct(self, 'ethtool -m {} 2> /dev/null'.format(port), quiet=True)
        elif flag == '-m hex on':
            return parse_to_dct(self, 'ethtool -m {} hex on 2> /dev/null'.format(port))
        elif flag == '--show-fec':
            cmd = 'ethtool --show-fec {}'.format(port)
            return shell_out_catch_errors(cmd)

    def lldpctl(self):
        output = shell_out_catch_errors('lldpctl -f keyvalue')
        return lldp_keyvalue_parser(output)

    def ptmctl(self):
        output = shell_out_catch_errors('ptmctl -d -j')
        return self.ptmctl_json_parser(output)

    def ptmctl_json_parser(self, output=None):
        if output and 'Unable to connect to PTMD' not in output:
            output_dct = json.loads(output)
            ptm_dct = {}
            for index in output_dct:
                if 'port' in output_dct[index] and output_dct[index]['port']:
                    port = output_dct[index]['port']
                    if 'exp nbr' in output_dct[index] and output_dct[index]['exp nbr']:
                        ptm_dct[port] = output_dct[index]['exp nbr']

            return ptm_dct
        else:
            return None

    def ip_link_show(self):
        return parse_to_dct(self, 'ip link show')

    def get_eeprom_list(self):
        eeprom_list = []
        pattern = re.compile(r'^swp(\d+).*$')
        filename = os.path.join(self.path_prefix, 'var/lib/cumulus/sfptab')
        f = open(filename, 'r')
        for swp in f.readlines():
            mo = pattern.search(swp)
            if mo:
                eeprom_list.append(mo.group(1))

        return eeprom_list

    def get_local_portwd_conf(self):
        if os.path.exists('/etc/cumulus/portwd.conf'):
            return '/etc/cumulus/portwd.conf'
        else:
            return None

    def get_porttab(self):
        if bcm_only(self):
            dut_install_bcm_dct(self, 'show portmap')
            bcm_show_portmap_dct = self.bcm_show_portmap_dct
        else:
            bcm_show_portmap_dct = {}

        debug('Collecting /v/l/c/porttab for {}'.format(self.hostname))
        filename = os.path.join(self.path_prefix, 'var/lib/cumulus/porttab')
        output = read_local_file(filename)
        if not output:
            critical('No ' + filename + '.  Cannot continue')
        parsed_porttab_dct = parse('cat /var/lib/cumulus/porttab', output)
        return convert_parsed_porttab(parsed_porttab_dct, bcm_show_portmap_dct)

    def get_pci_name(self):
        return "$(/usr/bin/mst status | grep pciconf | /usr/bin/awk '{{print $1}}')"

    def format_mlxlink_cmd(self, base, port):
        return base.format(
            mlxlink_cmd='/usr/bin/mlxlink',
            pci_name=self.get_pci_name(),
            # Extra space needs to be here for cl_support parsing
            extra_flag='--show_device ',
            mlx_port=port,
            post_cmd='2> /dev/null',
        )

    def get_mlxlink_data(self, cmd):
        output = shell_out_catch_errors(cmd)
        return parse(cmd, output)

    def init_mlx_sdk_handle(self):
        debug("opening sdk")
        rc, handle = sx_api_open(None)
        debug("sx_api_open handle:0x%x , rc %d " % (handle, rc))
        if rc != SX_STATUS_SUCCESS:
            error("Failed to open api handle. Please check that SDK is running.")
            sys.exit(errno.EACCES)
        return handle

    def deinit_mlx_sdk_handle(self, handle):
        sx_api_close(handle)

    def install_linux_to_mlx(self):
        pt = cumulus.porttab.porttab(use_hw=False)
        for linux_port in pt.get_linux_ports():
            self.linux_to_sdk[linux_port]=pt.linux2mlx(linux_port)
        debug('Port->SDK mapping: {}'.format(self.linux_to_sdk))

    def install_pcs_data(self, port_list, clear=False):
        handle = self.init_mlx_sdk_handle()

        for port in port_list:
            if port in self.linux_to_sdk and self.linux_to_sdk[port]:
                self.pcs_counters_dct[port] = self.get_pcs_data(handle, log_port=self.linux_to_sdk[port], clear=clear)
                debug('Added pcs counters for port: {}'.format(port))
                debug('Counters: {}'.format(self.pcs_counters_dct[port]))
            else:
                warn('Port {} is not present on the switch'.format(port))

        self.deinit_mlx_sdk_handle(handle)

    def get_pcs_data(self, handle, log_port=0, clear=0, link_side=0):
        pcs_port_data = {}
        cmds = {0: SX_ACCESS_CMD_READ, 1: SX_ACCESS_CMD_READ_CLEAR}
        if clear in cmds:
            cmd = cmds[clear]
        else:
            cmd = cmds[0]
        # SX_PORT_CNTR_GRP_PHY_LAYER_INTERNAL_LINK is supposed to be supported on Spec3 only, but now it just returns
        # 'unsupported cmd' 13 on existing Spec3 platforms, so we can safely skip it
        cntr_grps = {0: SX_PORT_CNTR_GRP_PHY_LAYER_STATISTICS, 1: SX_PORT_CNTR_GRP_PHY_LAYER_INTERNAL_LINK}
        cntr_grp = cntr_grps[0]
        if not log_port:
            warn('Logical port {} not valid.  Skipping'.format(log_port))
            return {}

        # Get all the data.  We only support phy stat counters now
        debug("Get PHY Statistics counters")
        cntr_p = new_sx_port_cntr_phy_layer_statistics_t_p()
        rc = sx_api_port_counter_phy_layer_statistics_get(handle, cmd, log_port, link_side, cntr_p)
        if rc != 0:
            error("PHY PCS Statistics counters read failed, return code: {}".format(rc))
            return {}

        if rc == SX_STATUS_SUCCESS:
            pcs_port_data["time_since_last_clear"] = cntr_p.time_since_last_clear
            pcs_port_data["phy_received_bits"] = cntr_p.phy_received_bits
            pcs_port_data["phy_symbol_errors"] = cntr_p.phy_symbol_errors
            pcs_port_data["phy_corrected_bits"] = cntr_p.phy_corrected_bits
            pcs_port_data["phy_raw_errors_lane0"] = cntr_p.phy_raw_errors_lane0
            pcs_port_data["phy_raw_errors_lane1"] = cntr_p.phy_raw_errors_lane1
            pcs_port_data["phy_raw_errors_lane2"] = cntr_p.phy_raw_errors_lane2
            pcs_port_data["phy_raw_errors_lane3"] = cntr_p.phy_raw_errors_lane3
            pcs_port_data["phy_raw_errors_lane4"] = cntr_p.phy_raw_errors_lane4
            pcs_port_data["phy_raw_errors_lane5"] = cntr_p.phy_raw_errors_lane5
            pcs_port_data["phy_raw_errors_lane6"] = cntr_p.phy_raw_errors_lane6
            pcs_port_data["phy_raw_errors_lane7"] = cntr_p.phy_raw_errors_lane7
            pcs_port_data["raw_ber_magnitude"] = cntr_p.raw_ber_magnitude
            pcs_port_data["raw_ber_coef"] = cntr_p.raw_ber_coef
            pcs_port_data["effective_ber_magnitude"] = cntr_p.effective_ber_magnitude
            pcs_port_data["effective_ber_coef"] = cntr_p.effective_ber_coef

        return pcs_port_data
