# Copyright (C) 2019-2020 Cumulus Networks,  Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import re
import os
import json
import sys
import fcntl
import inspect
import curses
import subprocess
import logging

from . import parse_port_module_data

from .parse_port_module_data import ETH_QSFP_CODE_MAP, ETH_SFP_CODE_MAP
from . import validate_data

comp_str = "compliance"

this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error


def colored(val, color):

    if os.environ.get("NO_TERMCOLOR", "FALSE") == "TRUE":
        return val

    # Check for support
    curses.initscr()
    support = curses.has_colors()
    curses.endwin()

    if not support:
        return val

    colors = dict(
        list(zip([
            'grey',
            'red',
            'green',
            'yellow',
            'blue',
            'magenta',
            'cyan',
            'white',
            ],
            list(range(30, 38))
            ))
        )

    color_exp = '\033\[(?:%s)m' % '|'.join(['%d' % v for v in colors.values()])
    reset_exp = '\033\[0m'
    reset = '\033[0m'
    fmt_str = '\033[%dm%s'

    val = re.sub(color_exp + '(.*?)' + reset_exp, r'\1', val)
    val = fmt_str % (colors[color], val)

    return val + reset


def get_load_funcs():
    return [o[1] for o in inspect.getmembers(parse_port_module_data, inspect.isfunction) if o[0].startswith("parse_")]


def concat_dict(x, y):
    z = x.copy()
    z.update(y)
    return z


def validate(arguments, port_type_map, db, data):

    debug('db: {}'.format(db))
    functions_list = [o[1] for o in inspect.getmembers(validate_data, inspect.isfunction)
                      if o[0].startswith("validate_")
                      and (comp_str in o[0]
                           or (arguments.get("--full", False) or "--full" not in arguments))]

    def val_wrapper(func, dat, val):

        key = func.__name__.split("validate_")[1]

        if dat == {}:
            return {key: {"result": "FAIL", "error": "No eeprom data received.", "data": "None"}}

        if val not in db:
            db[val] = [{"compliance_codes": [val], "primary_code": val}]
            debug('val_wrapper: Set db[{}] to {}'.format(val, db[val]))

        res = {key: func(dat, db[val], key)}
        if key != "meta" and res[key]:
            res[key].update({"raw_data": dat.get(key, None)})
        return res

    debug('port_type_map: {}'.format(port_type_map))
    return {dut: concat_dict({port: dict([i for sl in [val_wrapper(func, port_results, port_type_map[dut][port]).items()
                                                       for func in functions_list]
                                          for i in sl])
                              for port, port_results in port_dct.items()
                              if port_type_map.get(dut, {}).get(port, {}) != {}},

                             {"platform_info": port_dct.get("platform_info", {})})
            for dut, port_dct in data.items()}


def show_results(val, level=0, dut="", port=""):

    rows, columns = subprocess.check_output(['/bin/stty', 'size']).split()
    left = int(columns) / 2
    right = int(columns) - left
    errors = ""

    for k, v in val.items():

        if not v:
            continue
        v.pop("meta", None)
        v.pop("platform_info", None)

        if "result" in v:
            if v["result"] == "FAIL":
                out = colored("FAIL", "red")
                errors += "\nFAIL {} on {} \n{} \n {} \n{} \n\n".format(port, dut,"-"*left, v["error"],"-"*left)
            elif v["result"] == "PASS":
                out = colored("PASS", "green")
            else:
                out = colored(v["result"], "yellow")

            print(str.ljust("{}{} [{}]: ".format("    "*level, k, str(v["data"])[:min(left - 20, len(str(v["data"])))]), left, " ") + str.rjust("{}".format(out), right, " "))
        else:
            print("{}{}".format("    "*level,k))
            errors += show_results(v, level + 1, k if dut == "" and port == "" else dut, k if dut != "" and port == "" else port)

    if level == 0:
        print(colored(errors, "red"))

    return errors


def save_results(arguments, val):

    # Export to JSON
    if arguments["--export"] is not None:
        fname = arguments["--export"]

        with open(fname, "w") as f:
            json.dump(val, f, indent=4)

    # Display
    if arguments["--export"] is None or ("--display" in arguments and arguments["--display"]):
        show_results(val)


def swp_sort_pattern(item):
    pattern = re.compile(r'''swp?(\d+)p?s?(\d+)?$''')

    r = pattern.match(item)
    if r:
        if r.group(2):
            num = float(r.group(1) + '.' + r.group(2))
        else:
            num = float(r.group(1))
    return num if r else '&'


def ethcode_pattern(a, b):
    def conv(x):
        x = x[0]
        if "1000base" in x:
            return 100
        if "4x" in x:
            return 100
        return int(float(x.split("-")[0].split("g")[0])*100)

    return conv(a) - conv(b)


def get_port_type_map_from_console(arguments, data):

    def show_picker(prompt, options):
        print("")
        print(prompt)
        print("")

        for i, opt in enumerate(options):
            print("[{}]: {}".format(colored(str(i + 1), "cyan"), opt[1]))

        print("")
        print("")

        while True:

            idx = input("Enter Number of Choice: ")

            try:
                idx = int(idx)
                return options[idx - 1][0]
            except Exception as e:
                print(colored("Invalid Selection.", "red"))

    port_type_map = {}
    port_type_cache = {}

    for dut, ports in data.items():
        for port in sorted(ports, key=lambda l: swp_sort_pattern(l)):
            data = ports[port]
            if data == {}:
                continue

            port_type_map.setdefault(dut, {})

            type_cache_key = '_'.join([data["vendor"]["name"],
                                       data["vendor"]["oui"],
                                       data["vendor"]["pn"],
                                       data["vendor"]["rev"]
                                       ])

            if type_cache_key and type_cache_key in port_type_cache:
                port_type_map[dut][port] = port_type_cache[type_cache_key]
                continue

            if port_type_map.get(dut, {}).get(port, None) is None:
                name = "Select Cable Type for {} on {} \nVendor: {} OUI: {} PN: {} Rev: {} Type: {}".format(
                        colored(dut, "red"),
                        colored(port, "red"),
                        colored(data["vendor"]["name"], "green"),
                        colored(data["vendor"]["oui"], "green"),
                        colored(data["vendor"]["pn"], "green"),
                        colored(data["vendor"]["rev"], "green"),
                        colored(data["identifier"], "green"))

                code_dict = ETH_QSFP_CODE_MAP if "QSFP" in data["identifier"] else ETH_SFP_CODE_MAP

                items = sorted(list(set([(k, d) for k, v in code_dict.items()
                                         for d in v])), ethcode_pattern)

                port_type_map[dut][port] = show_picker(name, items)
                if type_cache_key and port_type_map[dut][port]:
                    port_type_cache[type_cache_key] = port_type_map[dut][port]

    return port_type_map


def dict_search(d, datum, path=[], prev=None):
    if type(d) is dict:
        for k in d.keys():
            dict_search(d[k], datum, path=(path + [k]), prev=d)
    else:
        for p in path[:-1]:
            if p not in datum:
                datum[p] = {}
            datum = datum[p]
        datum[path[-1]] = prev[path[-1]] + datum.get(path[-1], [])


def get_port_type_map(arguments, data):

    data = {k: {p: pv for p, pv in v.items()
                if "compliance_codes" in pv}
            for k, v in data.items()
            if len({p: pv for p, pv in v.items()
                    if "compliance_codes" in pv}.items()) > 0}

    if data == {}:
        print("No valid module(s) could be read! Aborting...")
        debug('No compliance codes.  ')
        quit()

    db = {}

    # Load data from file if specified
    if arguments["--type-map"] is not None:
        c_fname = arguments["--type-map"]

        with open(c_fname, "r") as f:
            cdata = json.load(f)

        port_type_map = {k: {port: cdata.get(k, {}).get(port, {})
                             for port, dat in v.items()}
                         for k, v in data.items()}

    else:
        # Poll user for missing data
        port_type_map = get_port_type_map_from_console(arguments, data)
        if 'localhost' in data and len(data['localhost']) > 1:
            exp = raw_input("Save [port -> compliance code] choices for later? [Y/n]")
            if exp.lower() != "n":
                fname = raw_input("Enter file location: [./port_compliance_map.json]")
                if fname == "":
                    fname = "port_compliance_map.json"
                with open(fname, "w") as f:
                    json.dump(port_type_map, f, indent=4)

    return port_type_map, db
