# Copyright (C) 2019 Cumulus Networks, Inc. all rights reserved
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import logging

this_log = logging.getLogger(__name__)

debug = this_log.debug
info = this_log.info
warn = this_log.warning
error = this_log.error


def get_all_items_matching_keys(group_of_dcts, key, default=None):
    def safe(c, k):
        try:
            return c[k]
        except Exception:
            return default

    if default is None:
        if isinstance(group_of_dcts, list):
            return [d[key] for d in group_of_dcts]
        elif isinstance(group_of_dcts, dict):
            return [v[key] for k, v in group_of_dcts.items()]

    else:
        return [safe(d, key) for d in group_of_dcts]


def make_entry(dat, val, func=None):
    def check_equal(dat, val):
        if type(dat) is not list and type(val) is list:
            return (True, None, dat) if dat in val else (False, "Port module values: {} not in expected values: {} for cable type".format(dat, list(set(val))), dat)
        return (True, None, dat) if dat == val else (False, "Port module value: {} does not equal expected value: {} for cable type".format(dat, val), dat)

    # Default to checking equal
    if func is None:
        func = check_equal
    res = func(dat, val)

    if res[0]:
        return {"result": "PASS", "data": res[2]}
    elif res[0] is None:
        return {"result": "SKIP", "data": res[2]}
    else:
        return {"result": "FAIL", "error": res[1], "data": res[2]}


def validate_meta(dat, val, key):
    return dat.get(key, {})


def validate_compliance_codes(dat, val, key):
    return make_entry(dat["primary_code"], val[0][key])


def validate_identifier(dat, val, key):
    if "QSFP" not in dat[key]:
        val_ident = ["SFP+"]
    else:
        # TODO  Fix this logic  4x25 is in SPEED, but not in primary_code (25g-cr...)
        if "100" in dat["primary_code"] or "25" in dat["primary_code"] or "50" in dat["primary_code"]:
            val_ident = ["QSFP28"]
        else:
            val_ident = ["QSFP+"]
    return make_entry(dat[key], val_ident)


def validate_connector(dat, val, key):
    connector = "".join(list(dat[key]))
    if dat["primary_code"] in ["100g-active", "40g-active", "10g-active"]:
        val_connector = ["none", "optical-pigtail", "no-separable-connector"]
    elif dat["primary_code"] in ["100g-active-cu", "100g-cr4", "40g-cr4", "10g-cr",
                                 "25g-cr-ca-s", "25g-cr-ca-n", "1000base-cx"]:
        val_connector = ["none", "copper-pigtail", "unknown", "no-separable-connector"]
    elif dat["primary_code"] in ["100g-sr4", "4x10g-sr", "40g-sr4", "100g-psm4", "40g-psm4"]:
        val_connector = ["mpo-1x12"]
    elif dat["primary_code"] in ["100g-lr4", "100g-er4", "100g-cwdm4fec", "100g-cwdm4", "100g-clr4",
                                 "100g-swdm4", "100g-pam4-bidi", "100g-dwdm2", "100g-wdm", "40g-er",
                                 "40g-swdm4", "40g-lr4", "10g-sr", "10g-lr", "10g-lrm", "10g-er",
                                 "1000base-sx", "1000base-lx"]:
        val_connector = ["lc"]
    elif dat["primary_code"] in ["10g-t", "1000base-t", "10g-tsr"]:
        val_connector = ["rj45"]
    else:
        val_connector = []
    return make_entry(connector, val_connector)


def validate_bitrate(dat, val, key):
    code = dat["primary_code"]
    if "10g" in code or "40g" in code:
        val_rate = range(10000, 11000, 100)
    elif "40g" in code or "25g" in code or "100g" in code:
        val_rate = range(25000, 26000, 100)
    elif "1000" in code:
        val_rate = range(1000, 3000, 100)
        val_rate += range(10000, 11000, 100)
    else:
        val_rate = []
    return make_entry(dat[key], val_rate)


def validate_vendor(dat, val, key):
    if dat[key] != {}:
        if len(dat[key]["name"]) != 0 and \
                len(dat[key]["oui"]) != 0 and \
                len(dat[key]["pn"]) != 0:
            return make_entry((True, None, "{} {} {} {}".format(dat[key]["name"],
                                                                dat[key]["oui"],
                                                                dat[key]["pn"],
                                                                dat[key]["rev"])), None, lambda x, y: x)

    return make_entry((False, "Incomplete Vendor Info", "{} {} {} {}".format(dat[key]["name"],
                                                        dat[key]["oui"],
                                                        dat[key]["pn"],
                                                        dat[key]["rev"])), None, lambda x, y: x)


def validate_DOM(dat, val, key):
    def check_dom(dat, val):

        if dat is None or "RX" not in dat: return (None, "No power data recieved.", "")

        for i in range(0, len(dat["RX"])):
            if dat["RX"][i] == 0.0:
                continue
            if dat["RX"][i] < 0.001:
                return (False, "{} mW below threshold 0.001 mW".format(dat["RX"][i]), ",".join([str(d) for d in dat["RX"]]) + " mW")
        return (True, None, ",".join([str(d) for d in dat["RX"]]) + " mW")
    return make_entry(dat[key], None, check_dom)


def validate_eye(dat, val, key):

    if "eye" not in dat: return make_entry((None, "No eye data from DUT", ""), None, lambda x, y: x)

    if dat["asic"] == "broadcom":
        up = 0
        down = 0
        left = 0
        right = 0
        try:
            up = min([t for t in get_all_items_matching_keys(dat[key], "U") if t != 0])
            down = min([t for t in get_all_items_matching_keys(dat[key], "D") if t != 0])
        except:
            pass
        try:
            right = min([t for t in get_all_items_matching_keys(dat[key], "R") if t != 0])
            left = min([t for t in get_all_items_matching_keys(dat[key], "L") if t != 0])
        except:
            pass

        if (up > 100 or up == 0) and (left > 200 or left == 0):
            return make_entry((True, None,  "Lowest: L: {} R: {} U: {} D: {}".format(left, right, up, down)), None, lambda x, y: x)
        else:
            return make_entry((False, "Lowest eye values below L/R: 200, U/D: 100 ", "L: {} R: {} U: {} D: {}".format(left, right, up, down)), None, lambda x, y: x)
    else:
        mV_test = 0
        try:
            mV_test = min([t for t in get_all_items_matching_keys(dat[key], "mV") if t != 0])
        except:
            pass
        grade_test = 0
        try:
            grade_test = min([t for t in get_all_items_matching_keys(dat[key], "grade") if t != 0])
        except:
            pass

        if (mV_test > 100 or mV_test == 0) and (grade_test > 4000 or grade_test == 0):
            return make_entry((True, None, "Lowest mV: {} grade: {}".format(mV_test, grade_test)), None, lambda x, y: x)
        else:
            return make_entry((False, "Lowest Eye values below height: 100mV / Grade: 4000", "mV: {} grade: {}".format(mV_test, grade_test)), None, lambda x, y: x)


def validate_link_status(dat, val, key):
    # Basic heruistic could be more tuned to final use case
    if key not in dat:
        return make_entry((None, "No link info", ""), None, lambda x, y: x)
    if dat[key] == "UP":
        return make_entry((True, None, "UP"), None, lambda x, y: x)
    else:
        return make_entry((False, "Link Down", "DOWN"), None, lambda x, y: x)


def validate_traffic(dat, val, key):
    if key not in dat:
        # No traffic was collected
        return

    failure = False
    if int(dat[key]['pkts']) > 0:
        # Traffic was generated. If not just skip this test, return nothing.
        for dutname in dat[key]:
            if dutname == 'pkts' or dutname == 'snake':
                continue
            for port in dat[key][dutname]:
                if ('HwIfInUcastPkts' in dat[key][dutname][port] and
                        'HwIfOutUcastPkts' in dat[key][dutname][port]):
                    if (dat[key][dutname][port]['HwIfInUcastPkts'] != dat[key]['pkts'] or
                            dat[key][dutname][port]['HwIfOutUcastPkts'] != dat[key]['pkts']):
                        failure = True
                else:
                    # No counter data, error
                    return make_entry((False, 'No Counter data collected for Dut: {} Port:{}'.format(dutname, port), ""),
                                      None, lambda x, y: x)

        if failure:
            return make_entry((False, 'Counters do not match traffic sent', int(dat[key]['pkts'])), None, lambda x, y: x)
        else:
            return make_entry((True, None, 'Counters match traffic sent'), None, lambda x, y: x)

